import math
from ._base import Material
from ..resources import Texture, TextureMap
from ..utils import logger, assert_type
from ..utils.color import Color
from ..utils.enums import ColorMode, VisibleSide


class MeshAbstractMaterial(Material):
    """Abstract mesh material.

    The abstract parent class for all mesh materials, defining their common properties.

    Parameters
    ----------
    color : Color
        The uniform color of the mesh (used depending on the ``color_mode``).
    color_mode : str | ColorMode
        The mode by which the mesh is coloured. Default 'auto'.
    map : TextureMap | Texture
        The texture map specifying the color at each texture coordinate. Optional.
    side : str | VisibleSide
        What side of the mesh is visible. Default "both".
    kwargs : Any
        Additional kwargs will be passed to the :class:`material base class
        <pygfx.Material>`.

    Notes
    -----
    The color format of the map is assumed to be sRGB. To use physical space
    instead, set the texture's colorspace property to "physical". To learn more
    about this, check out the :ref:`colorspace documentation <colorspaces>`

    The direction of a face is determined using Counter-clockwise (CCW) winding;
    i.e., if the fingers of your curled hand match the direction in which the
    face's vertices are defined then your thumb points into the "front"
    direction of the face. If this is not the case for your mesh, adjust its
    geometry (using e.g. ``np.fliplr()`` on ``geometry.indices``).

    """

    uniform_type = dict(
        Material.uniform_type,
        color="4xf4",
        wireframe="f4",
    )

    def __init__(
        self,
        color="#fff",
        color_mode="auto",
        map=None,
        side="both",
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.color = color
        self.color_mode = color_mode
        self.map = map
        self.side = side

    @property
    def color(self):
        """The uniform color of the mesh.
        This value is ignored if a texture map is used.
        """
        return Color(self.uniform_buffer.data["color"])

    @color.setter
    def color(self, color):
        color = Color(color)
        self.uniform_buffer.data["color"] = color
        self.uniform_buffer.update_full()

    @property
    def color_mode(self):
        """The way that color is applied to the mesh.

        See :obj:`pygfx.utils.enums.ColorMode`:
        """
        # todo: does 'auto' take the presence of texcoords into account?
        return self._store.color_mode

    @color_mode.setter
    def color_mode(self, value):
        value = value or "auto"
        if value not in ColorMode:
            raise ValueError(
                f"MeshMaterial.color_mode must be a string in {ColorMode}, not {value!r}"
            )
        self._store.color_mode = value

    @property
    def vertex_colors(self):
        return self.color_mode == ColorMode.vertex

    @vertex_colors.setter
    def vertex_colors(self, value):
        raise DeprecationWarning(
            "vertex_colors is deprecated, use ``color_mode='vertex'``"
        )

    @property
    def map(self):
        """The texture map specifying the color for each texture coordinate.
        The dimensionality of the map can be 1D, 2D or 3D, but should
        match the number of columns in the geometry's texcoords.

        The colors in the map are assumed to be in sRGB space. To use
        physical space instead, set the texture's colorspace property
        to "physical".
        """
        return self._store.map

    @map.setter
    def map(self, map):
        assert_type("map", map, None, Texture, TextureMap)
        if isinstance(map, Texture):
            map = TextureMap(map)
        self._store.map = map

    @property
    def side(self):
        """Defines which side of faces will be rendered.

        See :obj:`pygfx.utils.enums.VisibleSide`:

        Which side of the mesh is the front is determined by the winding of the faces.
        Counter-clockwise (CCW) winding is assumed. If this is not the case,
        adjust your geometry (using e.g. ``np.fliplr()`` on ``geometry.indices``).
        """
        return self._store.side

    @side.setter
    def side(self, value):
        value = (value or "both").lower()
        if value not in VisibleSide:
            raise ValueError(
                f"MeshMaterial.side must be a string in {VisibleSide}, not {value!r}"
            )
        self._store.side = value


class MeshBasicMaterial(MeshAbstractMaterial):
    """Basic mesh material.

    A material for drawing geometries in a simple shaded (flat or wireframe)
    way. This material is not affected by lights.

    Parameters
    ----------
    env_map : Texture
        The environment map.
    wireframe : bool
        If True, render geometry as a wireframe, i.e., only render edges.
    wireframe_thickness : int
        The thickness of a rendered edge in screen pixels.
    flat_shading : bool
        If True, the shader will ignore the geometry's normal data and instead
        use face normals during lighting calculations.
    reflectivity : float
        How much the environment map affects the surface. also see ``env_combine_mode``.
        The default value is 1 and the valid range is between 0 (no reflections) and 1 (full reflections).
    refraction_ratio : float
        The index of refraction (IOR) of air (approximately 1) divided by the index of refraction of the material.
        It is used with ``env_mapping_mode`` set to "REFRACTION".
    env_combine_mode: str
        How the environment map affects the surface.
        The default value is "MULTIPLY" and the valid values are "MULTIPLY", "MIX", and "ADD".
    env_mapping_mode : str
        The environment mapping mode.
        The default value is "CUBE-REFLECTION" and the valid values are "CUBE-REFLECTION" and "CUBE-REFRACTION".
    kwargs : Any
        Additional kwargs will be passed to the :class:`base class <pygfx.MeshAbstractMaterial>`.
    """

    uniform_type = dict(
        MeshAbstractMaterial.uniform_type,
        reflectivity="f4",
        refraction_ratio="f4",
        light_map_intensity="f4",
        ao_map_intensity="f4",
    )

    def __init__(
        self,
        env_map=None,
        wireframe=False,
        wireframe_thickness=1,
        flat_shading=False,
        reflectivity=1.0,
        refraction_ratio=0.98,
        env_combine_mode="MULTIPLY",
        env_mapping_mode="CUBE-REFLECTION",
        light_map=None,
        ao_map=None,
        specular_map=None,
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.wireframe = wireframe
        self.wireframe_thickness = wireframe_thickness
        self.flat_shading = flat_shading

        self.env_map = env_map
        self.env_combine_mode = env_combine_mode
        self.env_mapping_mode = env_mapping_mode
        self.reflectivity = reflectivity
        self.refraction_ratio = refraction_ratio

        self.light_map = light_map
        self.light_map_intensity = 1.0

        self.ao_map = ao_map
        self.ao_map_intensity = 1.0

        self.specular_map = specular_map

    @property
    def env_map(self):
        """The environment map."""
        return self._store.env_map

    @env_map.setter
    def env_map(self, env_map):
        assert env_map is None or isinstance(env_map, (Texture, TextureMap))
        if isinstance(env_map, Texture):
            env_map = TextureMap(env_map)
        self._store.env_map = env_map

    @property
    def wireframe(self):
        """Render geometry as a wireframe. Default is False (i.e. render as polygons)."""
        return self._store.wireframe

    @wireframe.setter
    def wireframe(self, value):
        is_wiremode = bool(value)
        self._store.wireframe = is_wiremode
        # Set uniform
        # We use a trick to make negative values indicate no-wireframe mode
        thickness = self.wireframe_thickness
        if is_wiremode:
            self.uniform_buffer.data["wireframe"] = thickness
        else:
            self.uniform_buffer.data["wireframe"] = -thickness
        self.uniform_buffer.update_full()

    @property
    def wireframe_thickness(self):
        """The thickness of the lines when rendering as a wireframe."""
        return abs(float(self.uniform_buffer.data["wireframe"])) or 1

    @wireframe_thickness.setter
    def wireframe_thickness(self, value):
        value = max(0.01, float(value))
        if self.uniform_buffer.data["wireframe"] > 0:
            self.uniform_buffer.data["wireframe"] = value
        else:
            self.uniform_buffer.data["wireframe"] = -value
        self.uniform_buffer.update_full()

    @property
    def flat_shading(self):
        """Whether the mesh is rendered with flat shading. When true,
        the shader will apply per-face surface normals, resulting in
        per-face lighting and a "pixelated", non-interpolated look,
        which can be useful to show the (size of) the triangle faces,
        or simply for the retro appearance. Note that the face normals
        are calculated from the vertex positions, ignoring the normal
        data in the geometry.
        """
        return self._store.flat_shading

    @flat_shading.setter
    def flat_shading(self, value: bool):
        self._store.flat_shading = bool(value)

    @property
    def reflectivity(self):
        """How much the environment map affects the surface. also see ``env_combine_mode``.
        The default value is 1 and the valid range is between 0 (no reflections) and 1 (full reflections).
        """
        return float(self.uniform_buffer.data["reflectivity"])

    @reflectivity.setter
    def reflectivity(self, value):
        self.uniform_buffer.data["reflectivity"] = value
        self.uniform_buffer.update_full()

    @property
    def refraction_ratio(self):
        """The index of refraction (IOR) of air (approximately 1) divided by the index of refraction of the material.
        It is used with ``env_mapping_mode`` set to "REFRACTION".
        """
        return float(self.uniform_buffer.data["refraction_ratio"])

    @refraction_ratio.setter
    def refraction_ratio(self, value):
        self.uniform_buffer.data["refraction_ratio"] = value
        self.uniform_buffer.update_full()

    @property
    def env_combine_mode(self):
        """How the environment map affects the surface.
        The default value is "MULTIPLY" and the valid values are "MULTIPLY", "MIX", and "ADD".
        """
        return self._store.env_combine_mode

    @env_combine_mode.setter
    def env_combine_mode(self, value):
        value = str(value).upper()
        if value in ("MULTIPLY", "MIX", "ADD"):
            self._store.env_combine_mode = value
        else:
            raise ValueError(f"Unexpected env_combine_mode: '{value}'")

    @property
    def env_mapping_mode(self):
        """The environment mapping mode.
        The default value is "REFLECTION" and the valid values are "CUBE-REFLECTION" and "CUBE-REFRACTION".
        """
        # todo: add support for other mapping modes,
        # "SPHERE-REFLECTION", "EQUIRECTANGULAR-REFLECTION", "EQUIRECTANGULAR-REFRACTION" etc.
        return self._store.env_mapping_mode

    @env_mapping_mode.setter
    def env_mapping_mode(self, value):
        value = str(value).upper()
        if value in ("CUBE-REFLECTION", "CUBE-REFRACTION"):
            self._store.env_mapping_mode = value
        else:
            raise ValueError(f"Unexpected env_mapping_mode: '{value}'")

    @property
    def light_map(self):
        """The light map to define pre-baked lighting (in srgb). Default is None.
        It usually requires a second set of texture coordinates."""
        return self._store.light_map

    @light_map.setter
    def light_map(self, map):
        assert_type("light_map", map, None, Texture, TextureMap)
        if isinstance(map, Texture):
            map = TextureMap(map)
        self._store.light_map = map

    @property
    def light_map_intensity(self):
        """Intensity of the baked light. Scaling occurs in the physical
        color space. Default is 1.0.
        """
        return float(self.uniform_buffer.data["light_map_intensity"])

    @light_map_intensity.setter
    def light_map_intensity(self, value):
        self.uniform_buffer.data["light_map_intensity"] = value
        self.uniform_buffer.update_full()

    @property
    def ao_map(self):
        """The red channel of this texture is used as the ambient occlusion map. Default is None.
        It usually requires a second set of texture coordinates."""
        return self._store.ao_map

    @ao_map.setter
    def ao_map(self, map):
        assert_type("ao_map", map, None, Texture, TextureMap)
        if isinstance(map, Texture):
            map = TextureMap(map)
        self._store.ao_map = map

    @property
    def ao_map_intensity(self):
        """Intensity of the ambient occlusion effect. Default is 1.0, zero is no occlusion effect."""
        return float(self.uniform_buffer.data["ao_map_intensity"])

    @ao_map_intensity.setter
    def ao_map_intensity(self, value):
        self.uniform_buffer.data["ao_map_intensity"] = value
        self.uniform_buffer.update_full()

    @property
    def specular_map(self):
        """The specular map. Default is None."""
        return self._store.specular_map

    @specular_map.setter
    def specular_map(self, map):
        assert_type("specular_map", map, None, Texture, TextureMap)
        if isinstance(map, Texture):
            map = TextureMap(map)
        self._store.specular_map = map


class MeshPhongMaterial(MeshBasicMaterial):
    """Phong mesh material.

    A material affected by light, diffuse and with specular
    highlights. This material uses the Blinn-Phong reflectance model.
    If the specular color is turned off, Lambertian shading is obtained.

    Parameters
    ----------
    shininess : int
        How shiny the specular highlight is; a higher value gives a sharper
        highlight.
    emissive : Color
        The emissive (light) color of the mesh. This color is added to the final
        color and is unaffected by lighting. The alpha channel of this color is
        ignored.
    specular : Color
        The specular (highlight) color of the mesh.
    kwargs : Any
        Additional kwargs will be passed to the :class:`base class
        <pygfx.MeshBasicMaterial>`.

    """

    # For reference:
    #
    # Lambertion shading, or Lambertian reflection, is a model to
    # calculate the diffuse component of a lit surface. Using this by
    # itself produces a matte look. All the below use a Lambertion term.
    #
    # Gouraud shading means doing the light-math in the vertex shader
    # and interpolating the final color over the face, often resulting
    # in a somewhat "interpolated" look. Back in the day this mattered
    # for performance, but it's silly now.
    #
    # Phong shading means interpolating the normals and doing the
    # light-math for each fragment.
    #
    # The Phong reflection model refers to the combination of ambient,
    # diffuse and specular lights, and the way that these are
    # calculated.
    #
    # The Blinn-Phong reflection model, also called the modified Phong
    # reflection model, is a tweak to how the reflection is calculated,
    # using a halfway factor, that was intended mostly as a performance
    # optimization, but apparently is a more accurate approximation of
    # how light behaves, or so they say.
    #
    # Flat shading refers to using the same color for the whole face.
    # This is what you get if the geometry has indices that do not share
    # vertices. But we can also obtain it by calculating the face normal
    # using derivatives of the world pos.

    uniform_type = dict(
        MeshBasicMaterial.uniform_type,
        emissive_color="4xf4",
        specular_color="4xf4",
        normal_scale="2xf4",
        shininess="f4",
        emissive_intensity="f4",
    )

    def __init__(
        self,
        shininess=30,
        emissive="#000",
        specular="#494949",  # as physical: #111, the default in ThreeJS
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.emissive = emissive
        self.shininess = shininess
        self.specular = specular

    @property
    def emissive(self):
        """The emissive (light) color of the mesh.
        This color is added to the final color and is unaffected by lighting.
        The alpha channel of this color is ignored.
        """
        return Color(self.uniform_buffer.data["emissive_color"])

    @emissive.setter
    def emissive(self, color):
        color = Color(color)
        self.uniform_buffer.data["emissive_color"] = color
        self.uniform_buffer.update_full()

    @property
    def specular(self):
        """The specular (highlight) color of the mesh."""

        return Color(self.uniform_buffer.data["specular_color"])

    @specular.setter
    def specular(self, color):
        color = Color(color)
        self.uniform_buffer.data["specular_color"] = color
        self.uniform_buffer.update_full()

    @property
    def shininess(self):
        """How shiny the specular highlight is; a higher value gives a sharper highlight.
        Default is 30.
        """
        return float(self.uniform_buffer.data["shininess"])

    @shininess.setter
    def shininess(self, value):
        self.uniform_buffer.data["shininess"] = value
        self.uniform_buffer.update_full()

    @property
    def emissive_map(self):
        """The emissive map color is modulated by the emissive color
        and the emissive intensity. If you have an emissive map, be
        sure to set the emissive color to something other than black.
        Note that both emissive color and emissive map are considered
        in srgb colorspace. Default None.
        """
        return self._store.emissive_map

    @emissive_map.setter
    def emissive_map(self, map):
        assert_type("emissive_map", map, None, Texture, TextureMap)
        if isinstance(map, Texture):
            map = TextureMap(map)
        self._store.emissive_map = map

    @property
    def emissive_intensity(self):
        """Intensity of the emissive light. Modulates the emissive color
        and emissive map. Default is 1.

        Note that the intensity is applied in the physical colorspace.
        You can think of it as scaling the number of photons. Therefore
        using an intensity of 0.5 is not the same as halving the
        emissive color, which is in srgb space.
        """
        return self.uniform_buffer.data["emissive_intensity"]

    @emissive_intensity.setter
    def emissive_intensity(self, value):
        self.uniform_buffer.data["emissive_intensity"] = value
        self.uniform_buffer.update_full()

    @property
    def normal_scale(self):
        """How much the normal map affects the material. This 2-tuple
        is multiplied with the normal_map's xy components (z is
        unaffected). Typical ranges are 0-1. Default is (1,1).
        """
        return tuple(self.uniform_buffer.data["normal_scale"])

    @normal_scale.setter
    def normal_scale(self, value):
        self.uniform_buffer.data["normal_scale"] = value
        self.uniform_buffer.update_range(0, 1)

    @property
    def normal_map(self):
        """The texture to create a normal map. Affects the surface
        normal for each pixel fragment and change the way the color is
        lit. Normal maps do not change the actual shape of the surface,
        only the lighting.
        """
        return self._store.normal_map

    @normal_map.setter
    def normal_map(self, map):
        assert_type("normal_map", map, None, Texture, TextureMap)
        if isinstance(map, Texture):
            map = TextureMap(map)
        self._store.normal_map = map


class MeshToonMaterial(MeshBasicMaterial):
    """
    A material implementing toon shading.

    Parameters
    ----------
    emissive : Color
        The emissive (light) color of the mesh. This color is added to the final
        color and is unaffected by lighting. The alpha channel of this color is
        ignored.

    gradient_map : Texture
        Gradient map for toon shading. The gradient map sampler method is always 'nearest'.

    kwargs : Any
        Additional kwargs will be passed to the :class:`base class
        <pygfx.MeshBasicMaterial>`.
    """

    uniform_type = dict(
        MeshBasicMaterial.uniform_type,
        emissive_color="4xf4",
        normal_scale="2xf4",
        emissive_intensity="f4",
    )

    def __init__(
        self,
        emissive="#000",
        gradient_map=None,
        emissive_intensity=1.0,
        emissive_map=None,
        normal_map=None,
        normal_scale=(1, 1),
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.emissive = emissive
        self.emissive_map = emissive_map
        self.emissive_intensity = emissive_intensity

        self.normal_map = normal_map
        self.normal_scale = normal_scale

        self.gradient_map = gradient_map

    @property
    def emissive(self):
        """The emissive (light) color of the mesh.
        This color is added to the final color and is unaffected by lighting.
        The alpha channel of this color is ignored.
        """
        return Color(self.uniform_buffer.data["emissive_color"])

    @emissive.setter
    def emissive(self, color):
        color = Color(color)
        self.uniform_buffer.data["emissive_color"] = color
        self.uniform_buffer.update_range(0, 1)

    @property
    def emissive_intensity(self):
        """Intensity of the emissive light. Modulates the emissive color
        and emissive map. Default is 1.

        Note that the intensity is applied in the physical colorspace.
        You can think of it as scaling the number of photons. Therefore
        using an intensity of 0.5 is not the same as halving the
        emissive color, which is in srgb space.
        """
        return float(self.uniform_buffer.data["emissive_intensity"])

    @emissive_intensity.setter
    def emissive_intensity(self, value):
        self.uniform_buffer.data["emissive_intensity"] = value
        self.uniform_buffer.update_range(0, 1)

    @property
    def normal_scale(self):
        """How much the normal map affects the material. This 2-tuple
        is multiplied with the normal_map's xy components (z is
        unaffected). Typical ranges are 0-1. Default is (1,1).
        """
        return tuple(self.uniform_buffer.data["normal_scale"])

    @normal_scale.setter
    def normal_scale(self, value):
        self.uniform_buffer.data["normal_scale"] = value
        self.uniform_buffer.update_range(0, 1)

    @property
    def normal_map(self):
        """The texture to create a normal map. Affects the surface
        normal for each pixel fragment and change the way the color is
        lit. Normal maps do not change the actual shape of the surface,
        only the lighting.
        """
        return self._store.normal_map

    @normal_map.setter
    def normal_map(self, map):
        assert_type("normal_map", map, None, Texture, TextureMap)
        if isinstance(map, Texture):
            map = TextureMap(map)
        self._store.normal_map = map

    @property
    def gradient_map(self):
        """Gradient map for toon shading.
        It's usually to set filter to 'nearest' for the gradient map.
        """
        return self._store.gradient_map

    @gradient_map.setter
    def gradient_map(self, map):
        assert_type("gradient_map", map, None, Texture, TextureMap)
        if isinstance(map, Texture):
            map = TextureMap(map, filter="nearest")
        self._store.gradient_map = map


class MeshNormalMaterial(MeshAbstractMaterial):
    """Color from Mesh normals.

    A material that maps the normal vectors to RGB colors.
    The ``flat_shading`` property can be used to show face normals.
    """


class MeshNormalLinesMaterial(MeshAbstractMaterial):
    """Render surface normals as lines.

    A material that shows surface normals as simple lines. The lines
    stick out from the vertices at the front faces by default.

    Parameters
    ----------
    line_length : float
        The length of the lines that indicate the normals, in local
        space. Set this to a negative value to make the lines stick out
        from the back faces.
    kwargs : Any
        Additional kwargs will be passed to the :class:`base class
        <pygfx.MeshBasicMaterial>`.

    """

    uniform_type = dict(
        MeshAbstractMaterial.uniform_type,
        line_length="f4",
    )

    def __init__(self, line_length=1.0, **kwargs):
        super().__init__(**kwargs)
        self.line_length = line_length

    def _wgpu_get_pick_info(self, pick_value):
        return {}  # No picking for normal lines

    @property
    def line_length(self):
        """The length of the lines that indicate the normals, in local
        space. Set this to a negative value to make the lines stick out
        from the back faces.
        """
        return float(self.uniform_buffer.data["line_length"])

    @line_length.setter
    def line_length(self, value):
        self.uniform_buffer.data["line_length"] = value
        self.uniform_buffer.update_full()


class MeshSliceMaterial(MeshAbstractMaterial):
    """Display a mesh slice.

    Parameters
    ----------
    plane : tuple
        The plane to slice at, represented with 4 floats ``(a, b, c, d)``, which
        make up the equation: ``ax + by + cz + d = 0`` The plane definition
        applies to the world space (of the scene).
    thickness : float
        The thickness of the line to draw the edge of the mesh in screen space
        (px).
    kwargs : Any
        Additional kwargs will be passed to the :class:`base class
        <pygfx.MeshBasicMaterial>`.

    """

    uniform_type = dict(
        MeshAbstractMaterial.uniform_type,
        plane="4xf4",
        thickness="f4",
    )

    def __init__(self, plane=(0, 0, 1, 0), thickness=2.0, **kwargs):
        super().__init__(**kwargs)
        self.plane = plane
        self.thickness = thickness

    @property
    def plane(self):
        """The plane to slice at, represented with 4 floats ``(a, b, c, d)``,
        which make up the equation: ``ax + by + cz + d = 0`` The plane
        definition applies to the world space (of the scene).
        """
        return tuple(self.uniform_buffer.data["plane"])

    @plane.setter
    def plane(self, plane):
        self.uniform_buffer.data["plane"] = plane
        self.uniform_buffer.update_full()

    @property
    def thickness(self):
        """The thickness of the line to draw the edge of the mesh."""
        return float(self.uniform_buffer.data["thickness"])

    @thickness.setter
    def thickness(self, thickness):
        self.uniform_buffer.data["thickness"] = thickness
        self.uniform_buffer.update_full()


class MeshStandardMaterial(MeshBasicMaterial):
    """PBR shaded material.

    A standard physically based material, applying PBR (Physically based rendering)
    using the Metallic-Roughness workflow.

    Parameters
    ----------
    emissive : Color
        The emissive color of the mesh. I.e. the color that the object emits
        even when not lit by a light source. This color is added to the final
        color and unaffected by lighting. The alpha channel is ignored.
    metalness : float
        How much the material looks like a metal. Non-metallic materials such as
        wood or stone use 0.0, metal use 1.0, with nothing (usually) in between.
        Default is 0.0. A value between 0.0 and 1.0 could be used for a rusty
        metal look. If metalness_map is also provided, both values are
        multiplied.
    roughness : float
        How rough the material is. 0.0 means a smooth mirror reflection, 1.0
        means fully diffuse. Default is 1.0. If roughness_map is also provided,
        both values are multiplied.
    kwargs : Any
        Additional kwargs will be passed to the :class:`base class
        <pygfx.MeshBasicMaterial>`.

    """

    # Physically based rendering (PBR) has recently become the standard
    # in many 3D applications, it use a physically correct model instead
    # of using approximations for the way in which light interacts with
    # a surface. Technical details of the approach can be found is this
    # paper from Disney (by Brent Burley):
    # https://media.disneyanimation.com/uploads/production/publication_asset/48/asset/s2012_pbs_disney_brdf_notes_v3.pdf

    uniform_type = dict(
        MeshBasicMaterial.uniform_type,
        emissive_color="4xf4",
        roughness="f4",
        metalness="f4",
        normal_scale="2xf4",
        emissive_intensity="f4",
        env_map_intensity="f4",
        env_map_max_mip_level="f4",
    )

    def __init__(
        self,
        *,
        emissive="#000",
        metalness=0.0,
        roughness=1.0,
        roughness_map=None,
        metalness_map=None,
        emissive_map=None,
        normal_map=None,
        env_map_intensity=1.0,
        normal_scale=(1, 1),
        emissive_intensity=1.0,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.emissive = emissive
        self.roughness = roughness
        self.metalness = metalness

        self.roughness_map = roughness_map
        self.metalness_map = metalness_map

        self.emissive_map = emissive_map
        self.emissive_intensity = emissive_intensity

        self.normal_map = normal_map
        self.normal_scale = normal_scale

        self.env_map_intensity = env_map_intensity

        # Note: there are more advanced properties to add, e.g. displacement_map, alpha_map

    @property
    def emissive(self):
        """The emissive color of the mesh. I.e. the color that the
        object emits even when not lit by a light source. This color
        is added to the final color and unaffected by lighting. The
        alpha channel is ignored.
        """
        return Color(self.uniform_buffer.data["emissive_color"])

    @emissive.setter
    def emissive(self, color):
        color = Color(color)
        self.uniform_buffer.data["emissive_color"] = color
        self.uniform_buffer.update_full()

    @property
    def emissive_map(self):
        """The emissive map color is modulated by the emissive color
        and the emissive intensity. If you have an emissive map, be
        sure to set the emissive color to something other than black.
        Note that both emissive color and emissive map are considered
        in srgb colorspace. Default None.
        """
        return self._store.emissive_map

    @emissive_map.setter
    def emissive_map(self, map):
        assert_type("emissive_map", map, None, Texture, TextureMap)
        if isinstance(map, Texture):
            map = TextureMap(map)
        self._store.emissive_map = map

    @property
    def emissive_intensity(self):
        """Intensity of the emissive light. Modulates the emissive color
        and emissive map. Default is 1.

        Note that the intensity is applied in the physical colorspace.
        You can think of it as scaling the number of photons. Therefore
        using an intensity of 0.5 is not the same as halving the
        emissive color, which is in srgb space.
        """
        return self.uniform_buffer.data["emissive_intensity"]

    @emissive_intensity.setter
    def emissive_intensity(self, value):
        self.uniform_buffer.data["emissive_intensity"] = value
        self.uniform_buffer.update_full()

    @property
    def metalness(self):
        """How much the material looks like a metal. Non-metallic materials
        such as wood or stone use 0.0, metal use 1.0, with nothing
        (usually) in between. Default is 0.0. A value between 0.0 and
        1.0 could be used for a rusty metal look. If metalness_map is
        also provided, both values are multiplied.
        """
        return float(self.uniform_buffer.data["metalness"])

    @metalness.setter
    def metalness(self, value):
        self.uniform_buffer.data["metalness"] = value
        self.uniform_buffer.update_full()

    @property
    def metalness_map(self):
        """The blue channel of this texture is used to alter the metalness of the material."""
        return self._store.metalness_map

    @metalness_map.setter
    def metalness_map(self, map):
        assert_type("metalness_map", map, None, Texture, TextureMap)
        if isinstance(map, Texture):
            map = TextureMap(map)
        self._store.metalness_map = map

    @property
    def roughness(self):
        """How rough the material is. 0.0 means a smooth mirror
        reflection, 1.0 means fully diffuse. Default is 1.0.
        If roughness_map is also provided, both values are multiplied.
        """
        return float(self.uniform_buffer.data["roughness"])

    @roughness.setter
    def roughness(self, value):
        self.uniform_buffer.data["roughness"] = value
        self.uniform_buffer.update_full()

    @property
    def roughness_map(self):
        """The green channel of this texture is used to alter the roughness of the material."""
        return self._store.roughness_map

    @roughness_map.setter
    def roughness_map(self, map):
        assert_type("roughness_map", map, None, Texture, TextureMap)
        if isinstance(map, Texture):
            map = TextureMap(map)
        self._store.roughness_map = map

    @property
    def normal_scale(self):
        """How much the normal map affects the material. This 2-tuple
        is multiplied with the normal_map's xy components (z is
        unaffected). Typical ranges are 0-1. Default is (1,1).
        """
        return tuple(self.uniform_buffer.data["normal_scale"])

    @normal_scale.setter
    def normal_scale(self, value):
        self.uniform_buffer.data["normal_scale"] = value
        self.uniform_buffer.update_full()

    @property
    def normal_map(self):
        """The texture to create a normal map. Affects the surface
        normal for each pixel fragment and change the way the color is
        lit. Normal maps do not change the actual shape of the surface,
        only the lighting.
        """
        return self._store.normal_map

    @normal_map.setter
    def normal_map(self, map):
        assert_type("normal_map", map, None, Texture, TextureMap)
        if isinstance(map, Texture):
            map = TextureMap(map)
        self._store.normal_map = map

    @property
    def env_map(self):
        """A texture that provides the environment map (in srgb colorspace). Default None.

        This makes the surroundings of the object be reflected on its
        surface. The given texture should have its ``generate_mipmaps`` set to
        True. Otherwise the roughness has no effect (as if its always zero).
        """
        # Note: to obtain a “physically correct” result, an advanced
        # mipmap technique is needed: PMREM (Prefiltered Mipmap Radiance
        # Environment Maps). We could (I think) add this technique in addition
        # to our normal mipmapping.
        return self._store.env_map

    @env_map.setter
    def env_map(self, map):
        assert_type("env_map", map, None, Texture, TextureMap)
        if isinstance(map, Texture):
            map = TextureMap(map)

        self._store.env_map = map

        if map is None:
            self.uniform_buffer.data["env_map_max_mip_level"] = 0
        else:
            if not map.texture.generate_mipmaps:
                logger.warning(
                    "The env_map texture must have generate_mipmaps=True in order for roughness to work."
                )
            width, height, _ = map.texture.size
            max_level = math.floor(math.log2(max(width, height))) + 1
            self.uniform_buffer.data["env_map_max_mip_level"] = float(max_level)
        self.uniform_buffer.update_full()

    @property
    def env_map_intensity(self):
        """Scales the effect of the environment map by multiplying its color.
        Note that this scaling occurs in the physical color space.
        """
        return float(self.uniform_buffer.data["env_map_intensity"])

    @env_map_intensity.setter
    def env_map_intensity(self, value):
        self.uniform_buffer.data["env_map_intensity"] = value
        self.uniform_buffer.update_full()


class MeshPhysicalMaterial(MeshStandardMaterial):
    """Physical mesh material.

    This is an extension of the MeshStandardMaterial,
    providing more advanced physically-based rendering properties.

    - **Clearcoat:** Some materials — like car paints, carbon fiber, and wet surfaces — require a clear,
    reflective layer on top of another layer that may be irregular or rough. Clearcoat approximates this effect,
    without the need for a separate transparent surface.

    - **Iridescence:** Allows to render the effect where hue varies depending on the viewing angle and
    illumination angle. This can be seen on soap bubbles, oil films, or on the wings of many insects.

    - **Anisotropy:** Ability to represent the anisotropic property of materials as observable with brushed metals.

    - **Sheen:** A soft, satin-like sheen on the surface, simulating the effect of a thin layer of fabric or a soft coating.


    Parameters
    ----------

    ior : float
        The index of refraction (IOR) of the material. Default is 1.5.
    specular : Color
        The specular (highlight) color of the mesh.
    clearcoat : float
        How much the material has a clearcoat layer. Default is 0.0.
    clearcoat_roughness : float
        How rough the clearcoat layer is. 0.0 means a smooth mirror
        reflection, 1.0 means fully diffuse. Default is 0.0.
    clearcoat_normal_scale : tuple
        How much the clearcoat normal map affects the material. This 2-tuple
        is multiplied with the clearcoat_normal_map's xy components (z is unaffected).
        Typical ranges are 0-1. Default is (1,1).
    iridescence : float
        The intensity of the iridescence layer, simulating RGB color shift based on the angle
        between the surface and the viewer, from 0.0 to 1.0. Default is 0.0.
    iridescence_ior : float
        The strength of the iridescence RGB color shift effect, represented by an index-of-refraction.
        Default is 1.3.
    iridescence_thickness_range : tuple
        The range of thickness for the iridescence effect, in nanometers. Default is (100, 400).
    anisotropy : float
        The anisotropy strength. Default is 0.0.
    anisotropy_rotation : float
        The rotation of the anisotropy in tangent, bitangent space, measured in radians counter-clockwise from the tangent.
        Default is 0.0.
    sheen : float
        The intensity of the sheen layer, simulating a soft, satin-like sheen on the surface.
        Default is 0.0.
    sheen_roughness : float
        The roughness of the sheen layer. from 0.0 to 1.0.
        Default is 1.0.
    sheen_color : Color
        The color of the sheen effect. Default is (0, 0, 0).

    kwargs : Any
        Additional kwargs will be passed to the :class:`base class
        <pygfx.MeshStandardMaterial>`.

    """

    # todo:
    # - Physically-based transparency
    # - Sheen
    #

    uniform_type = dict(
        MeshStandardMaterial.uniform_type,
        ior="f4",
        specular_color="4xf4",
        specular_intensity="f4",
        clearcoat="f4",
        clearcoat_roughness="f4",
        clearcoat_normal_scale="2xf4",
        iridescence="f4",
        iridescence_ior="f4",
        iridescence_thickness_range="2xf4",
        anisotropy_vector="2xf4",
        sheen="f4",
        sheen_color="4xf4",
        sheen_roughness="f4",
    )

    def __init__(
        self,
        ior=1.5,
        specular="#fff",
        specular_map=None,
        specular_intensity=1.0,
        specular_intensity_map=None,
        clearcoat=0.0,
        clearcoat_map=None,
        clearcoat_normal_map=None,
        clearcoat_normal_scale=(1, 1),
        clearcoat_roughness=0.0,
        clearcoat_roughness_map=None,
        iridescence=0.0,
        iridescence_map=None,
        iridescence_ior=1.3,
        iridescence_thickness_range=(100, 400),
        iridescence_thickness_map=None,
        anisotropy=0.0,
        anisotropy_map=None,
        anisotropy_rotation=0.0,
        sheen=0.0,
        sheen_roughness=1.0,
        sheen_roughness_map=None,
        sheen_color=(0, 0, 0),
        sheen_color_map=None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.ior = ior
        self.specular = specular
        self.specular_map = specular_map
        self.specular_intensity = specular_intensity
        self.specular_intensity_map = specular_intensity_map

        self.clearcoat = clearcoat
        self.clearcoat_map = clearcoat_map
        self.clearcoat_normal_map = clearcoat_normal_map
        self.clearcoat_normal_scale = clearcoat_normal_scale
        self.clearcoat_roughness = clearcoat_roughness
        self.clearcoat_roughness_map = clearcoat_roughness_map

        self.iridescence = iridescence
        self.iridescence_map = iridescence_map
        self.iridescence_ior = iridescence_ior
        self.iridescence_thickness_range = iridescence_thickness_range
        self.iridescence_thickness_map = iridescence_thickness_map

        self._anisotropy = anisotropy
        self._anisotropy_rotation = anisotropy_rotation
        self._update_anisotropy_vector()
        self.anisotropy_map = anisotropy_map

        self.sheen = sheen
        self.sheen_roughness = sheen_roughness
        self.sheen_roughness_map = sheen_roughness_map
        self.sheen_color = sheen_color
        self.sheen_color_map = sheen_color_map

    @property
    def ior(self):
        """The index of refraction (IOR) of the material. Default is 1.5."""
        return float(self.uniform_buffer.data["ior"])

    @ior.setter
    def ior(self, value):
        self.uniform_buffer.data["ior"] = value
        self.uniform_buffer.update_full()

    @property
    def specular(self):
        """The specular (highlight) color of the mesh."""
        return Color(self.uniform_buffer.data["specular_color"])

    @specular.setter
    def specular(self, color):
        color = Color(color)
        self.uniform_buffer.data["specular_color"] = color
        self.uniform_buffer.update_full()

    @property
    def specular_map(self):
        """The specular map. Default is None."""
        return self._store.specular_map

    @specular_map.setter
    def specular_map(self, map):
        assert_type("specular_map", map, None, Texture, TextureMap)
        if isinstance(map, Texture):
            map = TextureMap(map)
        self._store.specular_map = map

    @property
    def specular_intensity(self):
        """Intensity of the specular highlight. Default is 1.0."""
        return float(self.uniform_buffer.data["specular_intensity"])

    @specular_intensity.setter
    def specular_intensity(self, value):
        self.uniform_buffer.data["specular_intensity"] = value
        self.uniform_buffer.update_full()

    @property
    def specular_intensity_map(self):
        """The red channel of this texture is used to alter the specular intensity of the material."""
        return self._store.specular_intensity_map

    @specular_intensity_map.setter
    def specular_intensity_map(self, map):
        assert_type("specular_intensity_map", map, None, Texture, TextureMap)
        if isinstance(map, Texture):
            map = TextureMap(map)
        self._store.specular_intensity_map = map

    @property
    def clearcoat(self):
        """How much the material has a clearcoat layer. Default is 0.0."""
        return float(self.uniform_buffer.data["clearcoat"])

    @clearcoat.setter
    def clearcoat(self, value):
        self.uniform_buffer.data["clearcoat"] = value
        self.uniform_buffer.update_full()

    @property
    def clearcoat_normal_scale(self):
        """How much the clearcoat normal map affects the material. This 2-tuple
        is multiplied with the clearcoat_normal_map's xy components (z is
        unaffected). Typical ranges are 0-1. Default is (1,1).
        """
        return tuple(self.uniform_buffer.data["clearcoat_normal_scale"])

    @clearcoat_normal_scale.setter
    def clearcoat_normal_scale(self, value):
        self.uniform_buffer.data["clearcoat_normal_scale"] = value
        self.uniform_buffer.update_full()

    @property
    def clearcoat_normal_map(self):
        """The texture to create a clearcoat normal map. Affects the surface
        normal for each pixel fragment and change the way the color is
        lit. Normal maps do not change the actual shape of the surface,
        only the lighting.
        """
        return self._store.clearcoat_normal_map

    @clearcoat_normal_map.setter
    def clearcoat_normal_map(self, map):
        assert_type("clearcoat_normal_map", map, None, Texture, TextureMap)
        if isinstance(map, Texture):
            map = TextureMap(map)
        self._store.clearcoat_normal_map = map

    @property
    def clearcoat_roughness(self):
        """How rough the clearcoat layer is. 0.0 means a smooth mirror
        reflection, 1.0 means fully diffuse. Default is 0.0.
        If clearcoat_roughness_map is also provided, both values are multiplied.
        """
        return float(self.uniform_buffer.data["clearcoat_roughness"])

    @clearcoat_roughness.setter
    def clearcoat_roughness(self, value):
        self.uniform_buffer.data["clearcoat_roughness"] = value
        self.uniform_buffer.update_full()

    @property
    def clearcoat_roughness_map(self):
        """The green channel of this texture is used to alter the clearcoat roughness of the material."""
        return self._store.clearcoat_roughness_map

    @clearcoat_roughness_map.setter
    def clearcoat_roughness_map(self, map):
        assert_type("clearcoat_roughness_map", map, None, Texture, TextureMap)
        if isinstance(map, Texture):
            map = TextureMap(map)
        self._store.clearcoat_roughness_map = map

    @property
    def iridescence(self):
        """The intensity of the iridescence layer, simulating RGB color shift based on the angle between the surface and the viewer, from 0.0 to 1.0.
        Default is 0.0.
        """
        return float(self.uniform_buffer.data["iridescence"])

    @iridescence.setter
    def iridescence(self, value):
        self.uniform_buffer.data["iridescence"] = value
        self.uniform_buffer.update_full()

    @property
    def iridescence_map(self):
        """The red channel of this texture is used to alter the iridescence of the material."""
        return self._store.iridescence_map

    @iridescence_map.setter
    def iridescence_map(self, map):
        assert_type("iridescence_map", map, None, Texture, TextureMap)
        if isinstance(map, Texture):
            map = TextureMap(map)
        self._store.iridescence_map = map

    @property
    def iridescence_ior(self):
        """
        The strength of the iridescence RGB color shift effect, represented by an index-of-refraction.
        Default is 1.3.
        """
        return float(self.uniform_buffer.data["iridescence_ior"])

    @iridescence_ior.setter
    def iridescence_ior(self, value):
        self.uniform_buffer.data["iridescence_ior"] = value
        self.uniform_buffer.update_full()

    @property
    def iridescence_thickness_range(self):
        """The range of the iridescence layer thickness. Default is (100, 400).
        If `.iridescence_thickness_map` is not defined, iridescence thickness will use only the second element of the given array.
        """
        return tuple(self.uniform_buffer.data["iridescence_thickness_range"])

    @iridescence_thickness_range.setter
    def iridescence_thickness_range(self, value):
        self.uniform_buffer.data["iridescence_thickness_range"] = value
        self.uniform_buffer.update_full()

    @property
    def iridescence_thickness_map(self):
        """A texture that defines the thickness of the iridescence layer, stored in the green channel.

        - `0.0` in the green channel will result in thickness equal to first element of the `iridescence_thickness_range`.
        - `1.0` in the green channel will result in thickness equal to second element of the `iridescence_thickness_range`.
        - Values in-between will linearly interpolate between the elements of the `iridescence_thickness_range`.
        """
        return self._store.iridescence_thickness_map

    @iridescence_thickness_map.setter
    def iridescence_thickness_map(self, map):
        assert_type("iridescence_thickness_map", map, None, Texture, TextureMap)
        if isinstance(map, Texture):
            map = TextureMap(map)
        self._store.iridescence_thickness_map = map

    @property
    def anisotropy(self):
        """The anisotropy strength of the material. Default is 0.0."""
        return self._anisotropy

    @anisotropy.setter
    def anisotropy(self, value):
        self._anisotropy = value
        self._update_anisotropy_vector()

    @property
    def anisotropy_rotation(self):
        """The rotation of the anisotropy in tangent, bitangent space, measured in radians counter-clockwise from the tangent.
        Default is 0.0."""
        return self._anisotropy_rotation

    @anisotropy_rotation.setter
    def anisotropy_rotation(self, value):
        self._anisotropy_rotation = value
        self._update_anisotropy_vector()

    def _update_anisotropy_vector(self):
        self.uniform_buffer.data["anisotropy_vector"] = (
            math.cos(self._anisotropy_rotation) * self._anisotropy,
            math.sin(self._anisotropy_rotation) * self._anisotropy,
        )
        self.uniform_buffer.update_full()

    @property
    def anisotropy_map(self):
        """The anisotropy map is used to define the anisotropy direction and strength.
        Red and green channels represent the anisotropy direction in [-1, 1] tangent, bitangent space, to be rotated by `.anisotropy_rotation`.
        The blue channel contains strength as [0, 1] to be multiplied by `.anisotropy`.
        """
        return self._store.anisotropy_map

    @anisotropy_map.setter
    def anisotropy_map(self, map):
        assert_type("anisotropy_map", map, None, Texture, TextureMap)
        if isinstance(map, Texture):
            map = TextureMap(map)
        self._store.anisotropy_map = map

    @property
    def sheen(self):
        """The intensity of the sheen layer, from 0.0 to 1.0. Default is 0.0"""
        return float(self.uniform_buffer.data["sheen"])

    @sheen.setter
    def sheen(self, value):
        self.uniform_buffer.data["sheen"] = value
        self.uniform_buffer.update_full()

    @property
    def sheen_roughness(self):
        """Roughness of the sheen layer, from 0.0 to 1.0. Default is 1.0."""
        return float(self.uniform_buffer.data["sheen_roughness"])

    @sheen_roughness.setter
    def sheen_roughness(self, value):
        self.uniform_buffer.data["sheen_roughness"] = value
        self.uniform_buffer.update_full()

    @property
    def sheen_roughness_map(self):
        """The alpha channel of this texture is multiplied against .sheenRoughness, for per-pixel control over sheen roughness.
        Default is None."""
        return self._store.sheen_roughness_map

    @sheen_roughness_map.setter
    def sheen_roughness_map(self, map):
        assert_type("sheen_roughness_map", map, None, Texture, TextureMap)
        if isinstance(map, Texture):
            map = TextureMap(map)
        self._store.sheen_roughness_map = map

    @property
    def sheen_color(self):
        """The sheen tint. Default is (0, 0, 0), black."""
        return Color(self.uniform_buffer.data["sheen_color"])

    @sheen_color.setter
    def sheen_color(self, color):
        color = Color(color)
        self.uniform_buffer.data["sheen_color"] = color
        self.uniform_buffer.update_full()

    @property
    def sheen_color_map(self):
        """The RGB channels of this texture are multiplied against .sheenColor, for per-pixel control over sheen tint.
        Default is None."""
        return self._store.sheen_color_map

    @sheen_color_map.setter
    def sheen_color_map(self, map):
        assert_type("sheen_color_map", map, None, Texture, TextureMap)
        if isinstance(map, Texture):
            map = TextureMap(map)
        self._store.sheen_color_map = map



from math import floor, ceil, log10
from typing import Literal, Callable

import numpy as np

from ._base import WorldObject
from ._more import Line, Points
from ._text import MultiText
from ..resources import Buffer
from ..materials import LineMaterial, PointsMarkerMaterial, TextMaterial
from ..utils.compgeo import get_visible_part_of_line_ndc


class Ruler(WorldObject):
    """An object to represent a ruler with tickmarks.

    Can be used to measure distances in a scene, or as an axis in a plot.
    The ruler object is a "compound" object; it has text, lines, and points as child objects.

    Usage:

    * Use the properties (most notably ``start_pos`` and ``end_pos``).
    * Call ``update()`` on each draw.

    Parameters
    ----------
    start_pos : tuple[float, float, float]
        The initial position of the start of the ruler.
    end_pos : tuple[float, float, float]
        The initial position of the end of the ruler.
    start_value : float
        The value (offset) at the start of the ruler.
    ticks : array-like or None
        A list of values, in world-space from the ruler's start_pos, where ticks should be drawn.
        If not given or None, this will be determined automatically.
    tick_format : str | Callable
        The format to represent ticks with, or a function to convert floats to str. Default "0.4g".
    tick_side : str
        Whether the tick texts are on the 'left' or 'right' of the line (from the p.o.v. of the ruler). Default left.
    tick_marker : str
        Any marker from the PointsMarkerMaterial. Sensible values are 'tick', 'tick_left', and 'tick_right'.
    tick_size : float
        The size of the tickmarks (in logical screen pixels). When using half ticks
        (tick_marker is 'tick_left' or 'tick_right', the effective size is halved). Default 8.
    ticks_at_end_points : bool
        Whether to draw ticks at the ruler's strat and end. Default False
    min_tick_distance: float
        The minimal distance between ticks in screen pixels, when using auto-ticks. Default 50.
    color : str | tuple[float, float, float, float]
        The color for the internal line, points and text objects. Default white.
    line_width : float
        The width of the line and tickmarks. Default 2.0.
    alpha_mode : str | None
        Override the default alpha mode for the line, points and text.
    render_queue : int | None
        Override the default render queue for the line, points and text.
    """

    def __init__(
        self,
        *,
        start_pos: tuple[float, float, float] = (0, 0, 0),
        end_pos: tuple[float, float, float] = (0, 0, 0),
        start_value: float = 0.0,
        ticks: list | None = None,
        tick_format: str | Callable = "0.4g",
        tick_side: Literal["left", "right"] = "left",
        tick_marker: Literal["tick", "tick_left", "tick_right"] = "tick",
        tick_size: float = 8.0,
        ticks_at_end_points=False,
        min_tick_distance: float = 50.0,
        color: str | tuple[float, float, float, float] = "#fff",
        line_width: float = 2.0,
        alpha_mode: str | None = None,
        render_queue: int | None = None,
    ):
        super().__init__()

        self.start_pos = start_pos
        self.end_pos = end_pos
        self.start_value = start_value
        self.ticks = ticks
        self._tick_size = float(tick_size)

        self.tick_format = tick_format
        self.tick_side = tick_side
        self.min_tick_distance = min_tick_distance
        self.ticks_at_end_points = ticks_at_end_points

        # Common kwargs for the materials of the child objects
        material_kwargs = dict(
            alpha_mode=alpha_mode,
            render_queue=render_queue,
        )

        # Create a line and points object, with a shared geometry.
        # The 'tick' marker is a hair-line (i.e. no body), so you only
        # see the edge of the marker; we can simply match the edge_width with the
        # thickness of the line!
        self._text = MultiText(
            material=TextMaterial(color=color, aa=True, **material_kwargs),
            screen_space=True,
        )
        geometry = self._text.geometry  # has .positions buffer
        geometry.sizes = Buffer(np.zeros(geometry.positions.nitems, "f4"))
        self._line = Line(
            geometry,
            LineMaterial(
                color=color, thickness=line_width, aa=False, **material_kwargs
            ),
        )
        self._points = Points(
            geometry,
            PointsMarkerMaterial(
                marker=tick_marker,
                color="red",
                edge_color=color,
                edge_width=line_width,
                size_mode="vertex",
                rotation_mode="curve",
                aa=False,
                **material_kwargs,
            ),
        )

        # NOTE: a potential improvement, for the marker material's support for ticks, would
        # be be able to make every nth tick  (longer/ticker).

        self.add(self._line, self._points, self._text)

        # todo: a material to draw proper tick marks

    def get_bounding_box(self):
        # A ruler does not have a bounding box by itself
        return None

    # -- Properties to easily access sub-objects

    @property
    def line(self):
        """The line object that shows the ruler's path."""
        return self._line

    @property
    def points(self):
        """The points object that shows the ruler's tickmarks."""
        return self._points

    @property
    def text(self):
        """The text object that shows the ruler's tick labels."""
        return self._text

    # Note: text should also be here eventually

    # -- Main properties

    @property
    def start_pos(self):
        """The start posision of the ruler, in model space.

        Note that the ruler's transform also affects positioning, but should
        generally not be used.
        """
        return self._start_pos

    @start_pos.setter
    def start_pos(self, pos):
        start_pos = np.array(pos, np.float64).reshape((3,))
        if start_pos.shape != (3,):
            raise ValueError("Ruler.start_pos must be a 3-element position.")
        self._start_pos = start_pos
        self._end_value = None

    @property
    def end_pos(self):
        """The end posision of the ruler, in model space."""
        return self._end_pos

    @end_pos.setter
    def end_pos(self, pos):
        end_pos = np.array(pos, np.float64).reshape((3,))
        if end_pos.shape != (3,):
            raise ValueError("Ruler.end_pos must be a 3-element position.")
        self._end_pos = end_pos
        self._end_value = None

    @property
    def start_value(self) -> float:
        """The value of the ruler at the start position (i.e. the offset)."""
        return self._start_value

    @start_value.setter
    def start_value(self, value: float):
        self._start_value = float(value)
        self._end_value = None

    @property
    def end_value(self) -> float:
        """The value at the end of the ruler (read-only)."""
        # Little caching mechanic. Props that affect the end_value set ._end_value to None
        if self._end_value is None:
            self._end_value = float(
                np.linalg.norm(self._end_pos - self._start_pos) + self._start_value
            )
        return self._end_value

    @property
    def ticks(self):
        """The ticks to show.

        * ``None`` for automatic ticks.
        * ``dict`` for explicit ticks. Values can be str or float. If float, they
          are formatted with ``tick_format``.
        * ``list`` / ``tuple`` / ``ndarray`` for a list of values.
        """
        return self._ticks

    @ticks.setter
    def ticks(self, ticks):
        if ticks is None:
            self._ticks = None
        elif isinstance(ticks, dict):
            # Copy the object, resolving keys and values to float and str
            self._ticks = {float(k): v for k, v in ticks.items()}
        elif isinstance(ticks, (tuple, list, np.ndarray)):
            self._ticks = [float(x) for x in ticks]
        else:
            raise TypeError("Ruler.ticks must be None, dict or list(like).")

    # -- Properties for tweaking

    @property
    def tick_format(self) -> str:
        """The format to display the tick values.

        * A string to use as the second arg in ``format()``, default "0.4g".
        * "km" to use mili/Kilo/Mega/Giga suffixes.
        * A function that maps (value, min_value, max_value) to a str.
        """
        return self._tick_format

    @tick_format.setter
    def tick_format(self, tick_format: str):
        if isinstance(tick_format, str):
            self._tick_format = str(tick_format)
        elif callable(tick_format):
            # Emperically check the given function
            try:
                r = tick_format(0, -1, 1)
            except TypeError as err:
                raise ValueError(
                    f"Incompatible tick_format function: {err!s}"
                ) from None
            if not isinstance(r, str):
                raise ValueError(
                    f"Incompatible tick_format function: it must return str, not {r.__class__.__name__}"
                )
            self._tick_format = tick_format

    @property
    def tick_side(self) -> str:
        """Whether the ticks are on the 'left' or 'right' of the line.

        Imagine standing on the start position, with the line in front of you.
        """
        return self._tick_side

    @tick_side.setter
    def tick_side(self, side: str):
        side = str(side).lower()
        if side in ("left", "right"):
            self._tick_side = side
        else:
            raise ValueError("Tick side must be 'left' or 'right'.")

    @property
    def tick_marker(self) -> str:
        """The marker used for the ticks.

        Alias for ``ruler.points.material.marker``.

        This can be any value in :obj:`pygfx.utils.enums.MarkerShape`.
        Sensible values include 'tick', 'tick_left', and 'tick_right'.
        """
        return self._points.material.marker

    @tick_marker.setter
    def tick_marker(self, marker: str):
        self._points.material.marker = marker

    @property
    def tick_size(self) -> float:
        """The size of the tick marker, i.e. the length of the little line segment."""
        return self._tick_size

    @tick_size.setter
    def tick_size(self, size: float):
        self._tick_size = float(size)

    @property
    def ticks_at_end_points(self) -> bool:
        """Whether to show tickmarks at the end-points."""
        return self._ticks_at_end_points

    @ticks_at_end_points.setter
    def ticks_at_end_points(self, value: bool):
        self._ticks_at_end_points = bool(value)

    @property
    def min_tick_distance(self) -> float:
        """The minimal distance between ticks in screen pixels, when using auto-ticks."""
        return self._min_tick_dist

    @min_tick_distance.setter
    def min_tick_distance(self, value: float):
        value = float(value)
        if value < 0.0:
            raise ValueError("tick distance must be larger than zero.")
        self._min_tick_dist = value

    @property
    def color(self):
        """The color of the ruler components.

        The getter is an alias for ``ruler.text.material.color``. Setting this
        value sets  ``ruler.text.material.color``, ``ruler.line.material.color``,
        and ``ruler.points.material.edge_color``.
        """
        return self._text.material.color

    @color.setter
    def color(self, color):
        self._text.material.color = color
        self._line.material.color = color
        self._points.material.edge_color = color

    @property
    def line_width(self) -> float:
        """The width of the line and tickmarks.

        The getter is an alias for ``ruler.line.material.thickness``. Setting this
        value sets  ``ruler.line.material.thickness`` and ``ruler.points.material.edge_width``.
        """
        return self._line.material.thickness

    @line_width.setter
    def line_width(self, line_width: float):
        self._line.material.thickness = line_width
        self._points.material.edge_width = line_width

    # -- Methods

    def update(self, camera, canvas_size):
        """Update the ruler.

        This must be called on every draw, right before rendering.

        Returns a dictionary with the following fields:

        * "tick_step": the calculated auto-tick-step.
        * "tick_values": the tick values that will be shown.
        """

        # Determine which part of the ruler is on screen and its length in screen pixels
        self._configure_for_screen(camera, canvas_size)

        # Anchor
        screen_vec = self._visible_part_screen_vec
        self._calculate_text_anchor(np.arctan2(screen_vec[1], screen_vec[0]))

        # Get the dict with visible ticks
        tick_auto_step = self._calculate_tick_step()
        visible_ticks = self._get_ticks_dict(tick_auto_step)

        # Update objects to show these ticks
        self._update_sub_objects(visible_ticks, tick_auto_step)

        # Return stats. This is a dict, so we can add more stuff later, if needed.
        return {
            "tick_step": tick_auto_step,
            "tick_values": list(visible_ticks.keys()),
        }

    def _configure_for_screen(self, camera, canvas_size):
        """Make the ruler aware of the camera and viewport size."""

        half_canvas_size = 0.5 * np.array(canvas_size, np.float64).reshape(1, 2)

        # Get ndc coords for begin and end pos. Use numpy broadcasting for performance and compactness.
        positions = np.column_stack(
            [
                np.vstack([self._start_pos, self._end_pos]),
                np.ones((2, 1), np.float64),
            ]
        )
        ndc_full = (
            camera.camera_matrix @ self.world.matrix @ positions[..., None]
        ).reshape(-1, 4)
        screen_full = (ndc_full[:, :2] / ndc_full[:, 3:4]) * half_canvas_size

        # Get what part of the line is visible
        t1, t2 = get_visible_part_of_line_ndc(ndc_full[0], ndc_full[1])

        # Get screen coords for visible selection.
        ndc_sel = np.array(
            [
                ndc_full[0] * (1 - t1) + ndc_full[1] * t1,
                ndc_full[0] * (1 - t2) + ndc_full[1] * t2,
            ]
        )
        screen_sel = (ndc_sel[:, :2] / ndc_sel[:, 3:4]) * half_canvas_size

        # Store values
        self._screen_vec = screen_full[1] - screen_full[0]
        start_value, end_value = self._start_value, self.end_value
        self._visible_part_coords = t1, t2
        self._visible_part_values = (
            start_value * (1.0 - t1) + end_value * t1,
            start_value * (1.0 - t2) + end_value * t2,
        )
        self._visible_part_screen_vec = screen_sel[1] - screen_sel[0]

    def _calculate_text_anchor(self, angle):
        """Calculate the best place to anchor the text labels.
        With this anchor, the text labels move smoothly without a jump
        to the other side, as the ruler is rotated.
        """
        if self._tick_side == "left":
            if abs(angle) <= 0.25 * np.pi:
                self._text_anchor = "bottom-center"
                self._text_anchor_offset = 5
            elif abs(angle) >= 0.75 * np.pi:
                self._text_anchor = "top-center"
                self._text_anchor_offset = 5
            elif angle < 0:
                self._text_anchor = "middle-left"
                self._text_anchor_offset = 10
            else:
                self._text_anchor = "middle-right"
                self._text_anchor_offset = 10
        else:
            if abs(angle) <= 0.25 * np.pi:
                self._text_anchor = "top-center"
                self._text_anchor_offset = 5
            elif abs(angle) >= 0.75 * np.pi:
                self._text_anchor = "bottom-center"
                self._text_anchor_offset = 5
            elif angle < 0:
                self._text_anchor = "middle-right"
                self._text_anchor_offset = 10
            else:
                self._text_anchor = "middle-left"
                self._text_anchor_offset = 10

    def _get_ticks_dict(self, tick_auto_step):
        """Get a tick-dict, derived from the user-given tick value,
        and constrained to the visual part of the ruler.
        """

        min_value, max_value = self._visible_part_values
        tick_format = self._tick_format

        def default_tick_format_func(val, min_val, max_val):
            if val == 0:
                return "0"
            else:
                return format(val, tick_format)

        # Select funtion to format the tick values
        if tick_format == "km":
            tick_format_func = kmg_tick_format_func
        elif isinstance(tick_format, str):
            tick_format_func = default_tick_format_func
        elif callable(tick_format):
            tick_format_func = tick_format
        else:  # Fallback
            tick_format = "0.4g"
            tick_format_func = default_tick_format_func

        # Triage based on what ticks are given
        ticks = self._ticks
        if ticks is None:
            # Auto-ticks
            tick_values = self._get_ticks_uniform(min_value, max_value, tick_auto_step)
            return {t: tick_format_func(t, min_value, max_value) for t in tick_values}

        elif isinstance(ticks, list):
            # A sequence of ticks
            return {
                t: tick_format_func(t, min_value, max_value)
                for t in ticks
                if min_value <= t <= max_value
            }

        else:  # isinstance(ticks, dict):
            # A dict with specified ticks, values can be str or float
            result = {}
            for t, v in ticks.items():
                if min_value <= t <= max_value:
                    if isinstance(v, (float, int)):
                        v = tick_format_func(v, min_value, max_value)
                    elif not isinstance(v, str):
                        v = str(v)
                    result[t] = v
            return result

    def _calculate_tick_step(self):
        """Calculate the tick step from the min_tick_distance."""

        min_tick_dist = self._min_tick_dist

        # Determine distances for visible selection
        world_dist = self._visible_part_values[1] - self._visible_part_values[0]
        screen_dist = float(np.linalg.norm(self._visible_part_screen_vec))

        # Fall back to full size if selection is zero. This way, the
        # value of step still makes sense, even when the ruler itself
        # is not on screen, and calling code may still use it to e.g.
        # configure a grid. Account for roundoff errors resulting in a nonzero value.
        if screen_dist < 1e-9:
            world_dist = self.end_value - self._start_value
            screen_dist = np.linalg.norm(self._screen_vec)

        # Determine step
        step = 0
        if world_dist > 0 and screen_dist > 0:
            scale = screen_dist / world_dist
            approx_step = min_tick_dist / scale
            power10 = 10 ** floor(log10(approx_step))
            for i in (1, 2, 2.5, 5, 10):
                maybe_step = i * power10
                if maybe_step > approx_step:
                    step = maybe_step
                    break

        return step

    def _get_ticks_uniform(self, min_value, max_value, step):
        """Get a uniformly distributed set of ticks."""

        if not step:
            return []

        first_tick = ceil(min_value / step) * step
        last_tick = floor(max_value / step) * step

        ticks = []
        t = first_tick
        while t <= last_tick:
            ticks.append(t)
            t += step

        return ticks

    def _update_sub_objects(self, ticks, tick_auto_step):
        """Update the sub-objects to show the given ticks."""
        assert isinstance(ticks, dict)

        tick_size = self._tick_size

        # Load config
        start_pos = self._start_pos
        end_pos = self._end_pos
        start_value = self._start_value
        end_value = self.end_value

        # Derive some more variables
        length = end_value - start_value
        vec = end_pos - start_pos
        if length:
            vec /= length

        # Get number of positions that we need
        n_positions = len(ticks) + 2

        # Apply anchor props
        if self._text._anchor != self._text_anchor:
            self._text.anchor = self._text_anchor
        if self._text._anchor_offset != self._text_anchor_offset:
            self._text.anchor_offset = self._text_anchor_offset

        # Get the geometry to provide us with enough slots. Keep sizes array in sync
        self._text.set_text_block_count(n_positions)
        positions_buffer = self._text.geometry.positions
        sizes_buffer = self._text.geometry.sizes
        if sizes_buffer.nitems != positions_buffer.nitems:
            sizes_buffer = self._text.geometry.sizes = Buffer(
                np.zeros(positions_buffer.nitems, "f4")
            )

        # Get arrays / list that we can write to
        positions = positions_buffer.data
        sizes = sizes_buffer.data
        text_blocks = self._text._text_blocks

        # Apply start point
        index = 0
        positions[index] = start_pos
        if self._ticks_at_end_points:
            sizes[index] = tick_size
            text_blocks[index].set_text(f"{self._start_value:0.4g}")
        else:
            sizes[index] = 0
            text_blocks[index].set_text("")

        # Collect ticks
        index += 1
        for value, text in ticks.items():
            pos = start_pos + vec * (value - start_value)
            positions[index] = pos
            sizes[index] = tick_size
            text_blocks[index].set_text(text)
            index += 1

        # Apply end point
        positions[index] = end_pos
        if self._ticks_at_end_points:
            sizes[index] = tick_size
            text_blocks[index].set_text(f"{end_value:0.4g}")
        else:
            sizes[index] = 0
            text_blocks[index].set_text("")

        # Hide the ticks close to the ends?
        if self._ticks_at_end_points and ticks:
            tick_values = list(ticks.keys())
            if abs(tick_values[0] - start_value) < 0.5 * tick_auto_step:
                text_blocks[1].set_text("")
            if abs(tick_values[-1] - end_value) < 0.5 * tick_auto_step:
                text_blocks[index - 1].set_text("")

        # Make sure that the subset is drawn, and that the buffers are synced
        positions[n_positions:] = (
            np.nan
        )  # prevent a partial join to be drawn in the line
        if positions_buffer.draw_range[1] != n_positions:
            positions_buffer.draw_range = 0, n_positions
        positions_buffer.update_full()
        sizes_buffer.update_full()


# ---- Helper functions


def kmg_tick_format_func(val, min_val, max_val):
    ref_value = max(abs(min_val), abs(max_val))
    if ref_value >= 10_000_000_000:
        mult, unit = 1 / 1_000_000_000, "G"
    elif ref_value >= 10_000_000:
        mult, unit = 1 / 1_000_000, "M"
    elif ref_value >= 10000:
        mult, unit = 1 / 1000, "K"
    elif ref_value < 0.0001:
        mult, unit = 1_000_000, "u"
    elif ref_value < 0.1:
        mult, unit = 1000, "m"
    else:
        mult, unit = 1, ""
    if val == 0:
        return "0"
    else:
        return format(mult * val, "0.4g") + unit



import wgpu  # only for flags/enums

from ....objects import Volume
from ....materials import VolumeSliceMaterial, VolumeRayMaterial
from ....resources import Texture

from .. import (
    register_wgpu_render_function,
    BaseShader,
    Binding,
    to_texture_format,
    GfxSampler,
    GfxTextureView,
    load_wgsl,
)

vertex_and_fragment = wgpu.ShaderStage.VERTEX | wgpu.ShaderStage.FRAGMENT


class BaseVolumeShader(BaseShader):
    def __init__(self, wobject):
        super().__init__(wobject)
        material = wobject.material
        geometry = wobject.geometry

        # Check grid
        if geometry.grid is None:
            raise ValueError("Volume.geometry must have a grid (texture).")
        elif not isinstance(geometry.grid, Texture):
            raise TypeError("Volume.geometry.grid must be a Texture")
        elif geometry.grid.dim != 3:
            raise TypeError("Volume.geometry.grid must a 3D texture (view)")

        # Set image format
        self["climcorrection"] = ""
        fmt = to_texture_format(geometry.grid.format)
        if "norm" in fmt or "float" in fmt:
            self["img_format"] = "f32"
            if "unorm" in fmt:
                self["climcorrection"] = " * 255.0"
            elif "snorm" in fmt:
                self["climcorrection"] = " * 255.0 - 128.0"
        elif "uint" in fmt:
            self["img_format"] = "u32"
        else:
            self["img_format"] = "i32"

        # Set gamma
        self["gamma"] = material.gamma

        # Channels
        self["img_nchannels"] = len(fmt) - len(fmt.lstrip("rgba"))

        # Colorspace
        self["colorspace"] = geometry.grid.colorspace
        if material.map is not None:
            self["colorspace"] = material.map.texture.colorspace

    def get_bindings(self, wobject, shared):
        geometry = wobject.geometry
        material = wobject.material

        bindings = [
            Binding("u_stdinfo", "buffer/uniform", shared.uniform_buffer),
            Binding("u_wobject", "buffer/uniform", wobject.uniform_buffer),
            Binding("u_material", "buffer/uniform", material.uniform_buffer),
        ]

        tex_view = GfxTextureView(geometry.grid)
        sampler = GfxSampler(material.interpolation, "clamp")
        bindings.append(Binding("s_img", "sampler/filtering", sampler, "FRAGMENT"))
        bindings.append(Binding("t_img", "texture/auto", tex_view, vertex_and_fragment))

        if material.map is not None:
            bindings.extend(self.define_img_colormap(material.map))

        bindings = {i: b for i, b in enumerate(bindings)}
        self.define_bindings(0, bindings)

        return {
            0: bindings,
        }


@register_wgpu_render_function(Volume, VolumeSliceMaterial)
class VolumeSliceShader(BaseVolumeShader):
    type = "render"

    def get_pipeline_info(self, wobject, shared):
        return {
            "primitive_topology": wgpu.PrimitiveTopology.triangle_list,
            "cull_mode": wgpu.CullMode.none,
        }

    def get_render_info(self, wobject, shared):
        return {
            "indices": (12, 1),
        }

    def get_code(self):
        return load_wgsl("volume_slice.wgsl")


@register_wgpu_render_function(Volume, VolumeRayMaterial)
class VolumeRayShader(BaseVolumeShader):
    type = "render"

    def get_bindings(self, wobject, shared):
        render_mode = wobject.material.render_mode

        # Fall back to MIP, because we've written our examples to use the plain VolumeRayMaterial for quite a while.
        # Deprecate / remove this after a few releases (now is june 2025)
        render_mode = render_mode or "mip"

        if not render_mode:
            raise RuntimeError(
                f"Invalid value for {wobject.material.__class__.__name__}.render_mode: {render_mode!r}. Use an appropriate volume material, e.g. VolumeMipMaterial or VolumeIsoMaterial."
            )
        self["mode"] = render_mode
        return super().get_bindings(wobject, shared)

    def get_pipeline_info(self, wobject, shared):
        return {
            "primitive_topology": wgpu.PrimitiveTopology.triangle_list,
            "cull_mode": wgpu.CullMode.front,  # the back planes are the ref
        }

    def get_render_info(self, wobject, shared):
        return {
            "indices": (36, 1),
        }

    def get_code(self):
        return load_wgsl("volume_ray.wgsl")



class AnimationClip:
    """AnimationClip is a collection of tracks that are played together."""

    def __init__(self, name="", duration=-1, tracks=None) -> None:
        self.name = name
        self.duration = duration
        self.tracks = tracks or []

        if self.duration < 0:
            self.duration = self.reset_duration()

    def reset_duration(self):
        """Reset the duration of the clip to the maximum duration of all tracks."""
        duration = 0
        for track in self.tracks:
            max_time = track.times[-1]
            duration = max(max_time, duration)
        return duration



"""
This module implements the Text and MultiText objects. This is where the text
rendering comes together. Most steps in the text rendering process come from
pygfx.utils.text, though most of the alignment is implemented here.

For details about the text rendering process, see pygfx/utils/text/README.md


## On text object, text blocks, text items, and text pieces

The text object maintains the text. It holds and manages the arrays for the
glyph locations and atlas indices (in its geometry). It is also the main
entrypoint for the user.

Text is divided into multiple blocks, typically one per line/paragraph, similar
to how Qt splits text in blocks in its editor component. This makes it easy to
edit one block, and then merely shift the other blocks to update the layout. In
the multi-text scenario, these blocks can be individually controlled and
positioned by the user.

Each text block again divides its text into multiple items. Each item is a unit
of text that is moved as a whole during layout. Each word typically becomes one
item. Each item is created from one or more text pieces which each can have a
different format (bold, italic, size).

## On layout

Layout is performed on each block, shifting the text items into position based
on text_align, anchor and direction. This positoning is done by offsetting the
item's array of glyph positions. The offset is applied when the item's positions
are copied into the glyph_data buffer.

The Text object also performs a high level layout by positioning the blocks.
The MultiText object does not do this, as the user is responsible for
positioning the blocks.
"""

from typing import List, Union

import numpy as np

from ..resources import Buffer
from ..utils import text as textmodule
from ..utils.bounds import Bounds
from ..utils.enums import TextAlign, TextAnchor
from ..geometries import Geometry
from ..materials import TextMaterial
from ._base import WorldObject


# Allow anchor to be written without a dash, for backward compat.
ANCHOR_ALIASES = {anchor.replace("-", ""): anchor for anchor in TextAnchor}
ANCHOR_ALIASES["center"] = ANCHOR_ALIASES["middle"] = "middle-center"

# We cache the extents of small whitespace strings to improve performance
WHITESPACE_EXTENTS = {}

# This is the dtype for per-glyph data. In the shader we can only use 32bit datatypes (maybe f16 in a future version).
# To safe memory, we combine the atlas_index and format mask into a single u32. Here they look like two uint16 but
# in the shader they're interpreted as a single word, and decomposed.
GLYPH_DTYPE = np.dtype(
    [
        ("pos", "f4", 2),
        ("size", "f4"),
        ("block_index", "u4"),
        ("atlas_index", "u2"),
        ("format", "u2"),  # bitmask encoding relative weight and more
    ]
)


def encode_format(relative_weight, relative_slant):
    """Encode format props into a u16"""
    # Weigth: -250 .. 500, steps of 50  -> 4 bits
    relative_weight = min(max(relative_weight, -250), 500)
    weight_4 = int((relative_weight + 250) / 50 + 0.4999)
    # weight = weight_4 * 50 - 250

    # Slant: -1.75 .. 2, steps of 0.25 -> 4 bits
    relative_slant = min(max(relative_slant, -1.75), 2)
    slant_4 = int((relative_slant + 1.75) * 4 + 0.4999)
    # slant = slant_4 / 4.0 - 1.75;

    # There'res 8 bits left for possible future use

    packed = (weight_4 << 12) + (slant_4 << 8)

    return packed


class TextEngine:
    """A small abstraction that allows subclasses of Text to use a different text engine."""

    def __init__(self):
        # The shaping step, returns (glyph_indices, positions, meta)
        self.shape_text = textmodule._shaper.shape_text_hb

        # The font selection step, returns (text, font_filename) tuples.
        self.select_font = textmodule.select_font

        # The glyph generation step. Returns an array with atlas indices.
        self.generate_glyph = textmodule.generate_glyph

    def get_ws_extent(self, ws, font, direction):
        """Get the extent of a piece of whitespace text. Results of small strings are cached."""
        key = (font.filename, direction)
        map = WHITESPACE_EXTENTS.setdefault(key, {})
        # Try on the full ws
        try:
            return map[ws]
        except KeyError:
            pass
        # Calculate extent on base components
        extent = 0
        for c in ws:
            try:
                extent += map[c]
            except KeyError:
                meta = self.shape_text(c, font.filename, direction)[2]
                map[c] = meta["extent"]
                extent += map[c]
        # Calculate relatively short ws strings (e.g. indentation in code)
        if len(ws) <= 20:
            map[ws] = extent
        return extent


class Text(WorldObject):
    """Render a piece of text.

    Can be used to render a piece of multi-paragraph text. Supports plain text
    as well as formatting with (a subset of) markdown.

    Updates to text are relatively efficient, because text is internally divided
    into text blocks (for lines/paragraphs). Use ``MultiText`` if you want to
    directly control the text blocks.

    Parameters
    ----------
    geometry : Geometry | str
        The geometry on which the ``Text`` will store the buffers required to
        render the text ("positions" and "glyph_data"). If omitted, a new
        geometry is created automatically. This argument can also be used to
        provide the text to render as a string (as an alias for the `text` arg).
    material : TextMaterial
        The text material to use. Can be omitted.
    text : str | list[str]
        The plain text to render (optional). The text is split in one TextBlock per line,
        unless a list is given, in which case each (str) item become a TextBlock.
    markdown : str | list[str]
        The text to render, formatted as markdown (optional).
        See ``set_markdown()`` for details on the supported formatting.
        The text is split in one TextBlock per line,
        unless a list is given, in which case each (str) item become a TextBlock.
    font_size : float
        The size of the font, in object coordinates or pixel screen coordinates,
        depending on the value of the ``screen_space`` property. Default 12.
    family : str, tuple
        The name(s) of the font to prefer.
    direction : str | None
        The text direction overload.
    screen_space : bool
        Whether the text is rendered in model space or in screen space (like a label).
        Default False (i.e. model-space).
    anchor : str | TextAnchor
        The position of the origin of the text. Default "middle-center".
    anchor_offset : float
        The offset (extra margin) for the 'top', 'bottom', 'left', and 'right' anchors.
    max_width : float
        The maximum width of the text. Words are wrapped if necessary. Default zero (no wrapping).
    line_height : float
        A factor to scale the distance between lines. A value of 1 means the
        "native" font's line distance. Default 1.2.
    paragraph_spacing : float
        An extra space between paragraphs. Default 0.
    text_align : str | TextAlign
        The horizontal alignment of the text. Can be "start",
        "end", "left", "right", "center", "justify" or "justify_all". Default
        "start". Text alignment is ignored for vertical text (direction 'ttb' or 'btt').
    text_align_last: str | TextAlign
        The horizontal alignment of the last line of the content element. Default "auto".
    """

    _text_engine = TextEngine()
    is_multi = False

    uniform_type = dict(
        WorldObject.uniform_type,
        rot_scale_transform="2x2xf4",
    )

    def __init__(
        self,
        geometry=None,
        material=None,
        *,
        text=None,
        markdown=None,
        font_size=12,
        family=None,
        direction=None,
        screen_space=False,
        anchor="middle-center",
        anchor_offset=0,
        max_width=0,
        line_height=1.2,
        paragraph_spacing=0,
        text_align="start",
        text_align_last="auto",
        visible: bool = True,
        render_order: float = 0,
        name: str = "",
    ):
        # Init as a world object
        if isinstance(geometry, (str, list)):
            text = text or geometry
            geometry = None
        if geometry is None:
            geometry = Geometry()
        if material is None:
            material = TextMaterial()
        super().__init__(
            geometry,
            material,
            visible=visible,
            render_order=render_order,
            name=name,
        )

        # --- check text input

        text_and_markdown = [i for i in (text, markdown) if i is not None]
        if len(text_and_markdown) > 1:
            raise TypeError("Either text or markdown must be given, not both.")

        # --- create buffers

        # The position of each text block
        self.geometry.positions = Buffer(np.zeros((8, 3), "f4"))
        # self.colors = Buffer(np.zeros((8,4), "f4"))-> we could later implement per-block colors

        # The per-glyph data, stored with a structured dtype
        self.geometry.glyph_data = Buffer(np.zeros(16, GLYPH_DTYPE))

        # --- init variables to help manage the glyph arrays

        # The number of allocated glyph slots.
        # This must be equal to _glyph_indices_top - len(_glyph_indices_gaps)
        self._glyph_count = 0
        # The index marking the maximum used in the arrays. All elements higher than _glyph_indices_top are free.
        self._glyph_indices_top = 0
        # Free slots that are not in the contiguous space at the end of the arrays.
        self._glyph_indices_gaps = set()

        # Track what blocks need an update. This set is shared with the TextBlock instances.
        self._text_blocks = []  # List of TextBlock instances. May not match length of positions.
        self._dirty_blocks = set()  # Set of ints (text_block.index)

        # --- other geomery-specific things

        self._aabb = np.zeros((2, 3), "f4")
        self._aabb_rev = None

        # --- set propss

        # Font props
        self.font_size = font_size
        self.family = family
        self.direction = direction
        self.screen_space = screen_space

        # Layout props
        self.anchor = anchor
        self.anchor_offset = anchor_offset
        self.max_width = max_width
        self.line_height = line_height
        self.paragraph_spacing = paragraph_spacing
        self.text_align = text_align
        self.text_align_last = text_align_last

        # --- set initial content

        if text is not None:
            self.set_text(text)
        elif markdown is not None:
            self.set_markdown(markdown)

    # --- private methods

    def _update_world_transform(self):
        # Update when the world transform has changed
        super()._update_world_transform()
        # When rendering in screen space, the world transform is used
        # to establish the point in the scene where the text is placed.
        # The only part of the local transform that is used is the
        # position. Therefore, we also keep a transform containing the
        # local rotation and scale, so that these can be applied to the
        # text in screen coordinates.
        # Note that this applies to the whole text, all text blocks rotate around
        # the text-object origin. To rotate text blocks around their own origin,
        # we should probably implement TextBlock.angle.
        matrix = self.local.matrix[:2, :2]
        self.uniform_buffer.data["rot_scale_transform"] = matrix.T

    def _update_object(self):
        super()._update_object()
        # Is called right before the object is drawn;

        dirty_blocks = self._dirty_blocks
        if not dirty_blocks:
            return  # early exit

        # Update blocks
        need_high_level_layout = False
        need_glyph_data_sync = False
        for index in dirty_blocks:
            try:
                block = self._text_blocks[index]
            except IndexError:
                continue  # block was removed after being marked dirty
            did_block_layout, did_new_glyph_data = block._update(self)
            need_high_level_layout |= did_block_layout
            need_glyph_data_sync |= did_new_glyph_data

        # Reset
        dirty_blocks.clear()

        # Update drawing range. Note that the "gaps" are still rendered.
        glyph_data_buf = self.geometry.glyph_data
        glyph_data_buf.draw_range = 0, self._glyph_indices_top

        # Higher-level layout
        if need_high_level_layout:
            self._layout_blocks()

        # Updating in full turns out to be more efficient than doing all the calls to update_indices
        if need_glyph_data_sync:
            glyph_data_buf.update_full()

    def _layout_blocks(self):
        total_rect = apply_high_level_layout(self)
        self._aabb = np.array(
            [
                (total_rect.left, total_rect.bottom, 0),
                (total_rect.right, total_rect.top, 0),
            ],
            "f4",
        )

    # --- font properties

    @property
    def font_size(self):
        """The font size.

        For text rendered in screen space (``screen_space`` property is set),
        the size is in logical pixels, and the object's local transform affects
        the final text size.

        For text rendered in world space (``screen_space`` property is *not*
        set), the size is in object coordinates, and the the object's
        world-transform affects the final text size.

        Note that font size is indicative only. Final glyph size further depends on the
        font family, as glyphs may be smaller (or larger) than the indicative
        size. Final glyph size may further vary based on additional formatting
        applied a particular subsection.

        """
        return self._font_size

    @font_size.setter
    def font_size(self, value):
        self._font_size = float(value)
        self._trigger_blocks_update(layout=True)  # only need re-layout

    @property
    def family(self):
        """The font family to use.

        The name(s) of the font to prefer. If multiple names are given, they are
        preferred in the given order. Characters that are not supported by any
        of the given fonts are rendered with the default font (from the Noto
        Sans collection).
        """
        return self._font_props.family

    @family.setter
    def family(self, family):
        if family is None:
            self._font_props = textmodule.FontProps()
        else:
            self._font_props = textmodule.FontProps(family)
        self._trigger_blocks_update(render_glyphs=True)

    @property
    def direction(self):
        """The font direction overload.

        If not set (i.e. set to None), the text direction is determined
        automatically based on the script, selecting either 'ltr' or 'rtl'
        (Asian scripts are rendered left-to-right by default).

        If set, valid values are 'ltr', 'rtl', 'ttb', or 'btt'.
        Can also specify two values, separated by a dash (e.g. 'ttb-rtl')
        to specify the word-direction and line-direction, respectively.
        """
        return self._direction

    @direction.setter
    def direction(self, direction):
        direction = direction or ""
        word_direction, _, line_direction = direction.partition("-")
        valid_directions = ("", "ltr", "rtl", "ttb", "btt")
        if not (
            word_direction in valid_directions and line_direction in valid_directions
        ):
            raise ValueError(
                "Text direction components must be None, 'ltr', 'rtl', 'ttb', or 'btt'."
            )
        if not word_direction and not line_direction:
            self._direction = ""
        elif word_direction and line_direction:
            self._direction = f"{word_direction}-{line_direction}"
        elif not line_direction:
            m = {"ltr": "ttb", "rtl": "ttb", "ttb": "rtl", "btt": "rtl"}
            line_direction = m[word_direction]
            self._direction = f"{word_direction}-{line_direction}"
        else:
            self._direction = f"{word_direction}-{line_direction}"

        self._trigger_blocks_update(render_glyphs=True)

    @property
    def screen_space(self):
        """Whether to render text in screen space.

        * False: Render in model-pace (making it part of the scene), and having a bounding box.
        * True: Render in screen-space (like a label).

        Note that in both cases, the local object's rotation and scale will
        still transform the text.
        """
        return self._store.screen_space

    @screen_space.setter
    def screen_space(self, screen_space):
        self._store.screen_space = bool(screen_space)
        self._trigger_blocks_update(layout=True)

    # --- layout properties

    @property
    def anchor(self):
        """The position of the origin of the text.

        Represented as a string representing the vertical and horizontal anchors,
        separated by a dash, e.g. "top-left" or "bottom-center".

        * Vertical values: "top", "middle", "baseline", "bottom".
        * Horizontal values: "left", "center", "right".

        See :obj:`pygfx.utils.enums.TextAnchor`:
        """
        return self._anchor

    @anchor.setter
    def anchor(self, anchor):
        # Init
        if anchor is None:
            anchor = "middle-center"
        elif not isinstance(anchor, str):
            raise TypeError("Text anchor must be str.")
        anchor = ANCHOR_ALIASES.get(anchor, anchor)
        anchor = anchor.lower().strip().replace("-", "_")
        if anchor not in TextAnchor.__fields__:
            raise ValueError(f"Text anchor must be one of {TextAnchor}. Got {anchor!r}")
        self._anchor = TextAnchor[anchor]
        self._trigger_blocks_update(layout=True)

    @property
    def anchor_offset(self):
        """The offset (extra margin) for the 'top', 'bottom', 'left', and 'right' anchors."""
        return self._anchor_offset

    @anchor_offset.setter
    def anchor_offset(self, value):
        self._anchor_offset = float(value)
        self._trigger_blocks_update(layout=True)

    @property
    def max_width(self):
        """The maximum width of the text.

        Text will wrap if beyond this limit. The coordinate system that this
        applies to is the same as for ``font_size``. Set to 0 for no wrap (default).

        For vertical text, this value indicates the maximum height.
        """
        return self._max_width

    @max_width.setter
    def max_width(self, width):
        self._max_width = float(width or 0)
        self._trigger_blocks_update(layout=True)

    @property
    def line_height(self):
        """The relative height of a line of text.

        Used to set the distance between lines. Default 1.2.
        For vertical text this also defines the distance between the (vertical) lines.
        """
        return self._line_height

    @line_height.setter
    def line_height(self, height):
        self._line_height = float(height or 1.2)
        self._trigger_blocks_update(layout=True)

    @property
    def paragraph_spacing(self):
        """The extra margin between two paragraphs.

        Measured in text units (like line height and font size).
        """
        return self._paragraph_spacing

    @paragraph_spacing.setter
    def paragraph_spacing(self, paragraph_spacing):
        self._paragraph_spacing = float(paragraph_spacing or 0)
        self._trigger_blocks_update(layout=True)

    @property
    def text_align(self):
        """Set the alignment of wrapped text. Default 'start'.

        See :obj:`pygfx.utils.enums.TextAlign`:

        Text alignment is ignored for vertical text (direction 'ttb' or 'btt').
        """
        return self._text_align

    @text_align.setter
    def text_align(self, align):
        if align is None:
            align = "start"
        if not isinstance(align, str):
            raise TypeError("Text align must be a None or str.")
        align = align.lower().replace("-", "_")
        if align not in TextAlign.__fields__:
            raise ValueError(f"Text align must be one of {TextAlign}. Got {align!r}.")
        if align == "auto":
            align = "start"
        self._text_align = TextAlign[align]
        self._trigger_blocks_update(layout=True)

    @property
    def text_align_last(self):
        """Set the alignment of the last line of text. Default "auto".

        See :obj:`pygfx.utils.enums.TextAlign`:

        Text alignment is ignored for vertical text (direction 'ttb' or 'btt').
        """
        return self._text_align_last

    @text_align_last.setter
    def text_align_last(self, align):
        if align is None:
            align = "auto"
        if not isinstance(align, str):
            raise TypeError("Text align_last must be a None or str.")
        align = align.lower().replace("-", "_")
        if align not in TextAlign.__fields__:
            raise ValueError(
                f"Text align_last must be one of {TextAlign}. Got {align!r}"
            )
        self._text_align_last = TextAlign[align]
        self._trigger_blocks_update(layout=True)

    # --- public methods

    def set_text(self, text: Union[str, List[str]]):
        """Set the full text.

        Each line (i.e. paragraph) results in one TextBlock, unless a list is given,
        in which case each (str) item become a TextBlock.

        On subsequent calls, blocks are re-used, and blocks that did not change
        have near-zero overhead (only lines/blocks that changed need updating).
        """
        if isinstance(text, str):
            str_per_block = text.splitlines()
        elif isinstance(text, list):
            str_per_block = text
        else:
            raise TypeError("Text text should be str.")
        self._ensure_text_block_count(len(str_per_block))
        for i, s in enumerate(str_per_block):
            # Note that setting the blocks text is fast if it did not change
            self._text_blocks[i].set_text(s)

        # Do a layout now, so the bounding box is up-to-date
        self._update_object()

    def set_markdown(self, text: Union[str, List[str]]):
        """Set the full text, formatted as markdown.

        The supported markdown features are:

        * ``**bold** and *italic* text`` is supported for words, word-parts,
          and (partial) sentences, but not multiple lines (formatting state does
          not cross line boundaries).
        * ``# h1``, ``## h2``, ``### h3``, etc.
        * ``* bullet points``.

        Each line (i.e. paragraph) results in one TextBlock, unless a list is given,
        in which case each (str) item become a TextBlock.

        On subsequent calls, blocks are re-used, and blocks that did not change
        have near-zero overhead (only lines/blocks that changed need updating).
        """
        if isinstance(text, str):
            str_per_block = text.splitlines()
        elif isinstance(text, list):
            str_per_block = text
        else:
            raise TypeError("Text markdown should be str.")
        self._ensure_text_block_count(len(str_per_block))
        for i, s in enumerate(str_per_block):
            self._text_blocks[i].set_markdown(s)

        # Do a layout now, so the bounding box is up-to-date
        self._update_object()

    def _get_bounds_from_geometry(self):
        screen_space = self._store["screen_space"]
        if not self._text_blocks:
            return None
        elif screen_space:
            # There is no sensible bounding box for text in screen space, except
            # for the anchor point. Although the point has no volume, it does
            # contribute to e.g. the scene's bounding box.
            return Bounds(np.zeros((2, 3), "f4"))
        else:
            # A bounding box makes sense, and we calculated it during layout,
            # because we're already shifting rects there.
            return Bounds(self._aabb)

    # --- block management

    def _trigger_blocks_update(self, layout=False, render_glyphs=False):
        for block in self._text_blocks:
            block._mark_dirty(layout=layout, render_glyphs=render_glyphs)

    def _ensure_text_block_count(self, n):
        """Allocate new buffer if necessary."""

        # Make sure the underlying buffers are large enough
        current_buffer_size = self.geometry.positions.nitems
        max_oversize = max(4 * n, 8)
        if current_buffer_size < n or current_buffer_size > max_oversize:
            new_size = 2 ** int(np.ceil(np.log2(max(n, 1))))
            new_size = max(8, new_size)
            self._allocate_block_buffers(new_size)

        if n == len(self._text_blocks):
            pass
        elif len(self._text_blocks) < n:
            # Add blocks
            while len(self._text_blocks) < n:
                block = TextBlock(len(self._text_blocks), self._dirty_blocks)
                self._text_blocks.append(block)
        else:
            # Remove blocks
            self.geometry.glyph_data.update_full()  # cleared blocks, means cleared glyph indices
            while len(self._text_blocks) > n:
                block = self._text_blocks.pop(-1)
                block._clear(self)

    def _allocate_block_buffers(self, n):
        """Allocate new buffers for text blocks with the given size."""
        smallest_n = min(n, len(self._text_blocks))
        new_positions = np.zeros((n, 3), "f4")
        new_positions[:smallest_n] = self.geometry.positions.data[:smallest_n]
        self.geometry.positions = Buffer(new_positions)

    # --- glyph array management

    def _glyphs_allocate(self, n):
        max_glyph_slots = self.geometry.glyph_data.nitems

        # Need larger buffer?
        n_free = max_glyph_slots - self._glyph_count
        if n > n_free:
            self._glyphs_create_new_buffers(n)

        # Allocate indices
        if not self._glyph_indices_gaps:
            # Contiguous: indices is a range
            assert self._glyph_indices_top == self._glyph_count
            indices = range(self._glyph_indices_top, self._glyph_indices_top + n)
            self._glyph_count += n
            self._glyph_indices_top += n
        else:
            # First use gaps ...
            indices = np.empty((n,), "u4")
            n_from_gap = min(n, len(self._glyph_indices_gaps))
            for i in range(n_from_gap):
                indices[i] = self._glyph_indices_gaps.pop()
            self._glyph_count += n_from_gap
            # Then use indices at the end
            n -= n_from_gap
            if n > 0:
                indices[n_from_gap:] = range(
                    self._glyph_indices_top, self._glyph_indices_top + n
                )
                self._glyph_count += n
                self._glyph_indices_top += n

        return indices

    def _glyphs_deallocate(self, indices):
        # These glyphs will still end up in the vertex shader,
        # but it will discard early by producing degeneate triangles.
        # Clear data, for sanity
        self.geometry.glyph_data.data[indices] = 0
        # self.glyph_data.update_indices(indices) -> update_full is called from _update_object when needed
        # Deallocate
        self._glyph_count -= len(indices)
        # Small optimization to avoid gaps
        if indices.min() == self._glyph_indices_top - len(indices):
            self._glyph_indices_top -= len(indices)
        else:
            self._glyph_indices_gaps.update(indices)
        # TODO: Reduce buffer sizes from the Text object, re-packing all items
        # # Maybe reduce buffer size
        # max_glyph_slots = self.glyph_data.nitems
        # if self._glyph_count < 0.25 * max_glyph_slots:
        #     self._glyphs_create_new_buffers()

    def _glyphs_create_new_buffers(self, extra_needed=0):
        assert extra_needed >= 0

        # Get new size
        need_size = self._glyph_indices_top + extra_needed
        new_size = 2 ** int(np.ceil(np.log2(need_size)))
        new_size = max(16, new_size)
        # Prepare new arrays
        glyph_data = np.zeros(new_size, GLYPH_DTYPE)
        # Copy data over
        n = self._glyph_indices_top
        glyph_data[:n] = self.geometry.glyph_data.data[:n]
        # Store
        self.geometry.glyph_data = Buffer(glyph_data)


class MultiText(Text):
    """Render multiple pieces of text.

    The ``MultiText`` object manages a collection of ``TextBlock`` objects,
    for which the text (or markdown) can be individually set, and which can
    be individually positioned. Each text block has the same layout support
    as the ``Text`` obect.

    Most properties are defined by the text object, i.e. shared across all
    ``TextBlock`` instances of that ``MultiText`` object. But this may change in
    the future to allow more flexibility.
    """

    is_multi = True

    def set_text_block_count(self, n):
        """Set the number of text blocks to n.

        Use this if you want to use text blocks directly and you know how many
        blocks you need beforehand. After this, get access to the blocks
        using ``get_text_block()``.
        """
        self._ensure_text_block_count(n)

    def get_text_block_count(self):
        """Get how many text blocks this MultiText object has."""
        return len(self._text_blocks)

    def create_text_block(self):
        """Create a text block and return it.

        The text block count is increased by one.
        """
        self._ensure_text_block_count(len(self._text_blocks) + 1)
        return self._text_blocks[-1]

    def create_text_blocks(self, n):
        """Create n text blocks and return as a list.

        The text block count is increased by n.
        """
        self._ensure_text_block_count(len(self._text_blocks) + n)
        return self._text_blocks[-n:]

    def get_text_block(self, index):
        """Get the TextBlock instance at the given index.

        The block's position is stored in ``text_object.geometry.positions.data[index]``.
        """
        return self._text_blocks[index]

    def _layout_blocks(self):
        total_rect = Rect()
        for block in self._text_blocks:
            total_rect.left = min(total_rect.left, block._rect.left)
            total_rect.right = max(total_rect.right, block._rect.right)
            total_rect.top = max(total_rect.top, block._rect.top)
            total_rect.bottom = min(total_rect.bottom, block._rect.bottom)

        self._aabb = np.array(
            [
                (total_rect.left, total_rect.bottom, 0),
                (total_rect.right, total_rect.top, 0),
            ],
            "f4",
        )

    def _get_bounds_from_geometry(self):
        screen_space = self._store["screen_space"]
        if screen_space:
            positions_buf = self.geometry.positions
            if not self._text_blocks:
                return None
            if self._aabb_rev == positions_buf.rev:
                return Bounds(self._aabb)
            aabb = None
            # Get positions and check expected shape
            positions = positions_buf.data[: len(self._text_blocks)]
            aabb = np.array([positions.min(axis=0), positions.max(axis=0)], "f4")
            # If positions contains xy, but not z, assume z=0
            if aabb.shape[1] == 2:
                aabb = np.column_stack([aabb, np.zeros((2, 1), "f4")])
            self._aabb = aabb
            self._aabb_rev = positions_buf.rev
            return Bounds(self._aabb)
        else:
            # A bounding box makes sense, and we calculated it during layout,
            # because we're already shifting rects there.
            return Bounds(self._aabb)


class TextBlock:
    """The TextBlock represents one block or paragraph of text.

    Users can obtain instances of this class from the ``MultiText`` object.

    Text blocks are positioned using an entry in ``text_object.geometry.positions``.
    The ``Text`` object uses text blocks internally to do efficient re-layout when text is updated.
    With the ``MultiText`` object the text blocks are positioned by the user or external code.
    """

    def __init__(self, index, dirty_blocks):
        self._index = index  # e.g. the index in geometry.positions
        self._dirty_blocks = dirty_blocks  # a set from the Text/MultiText object

        self._input = None
        self._need_layout = False
        self._need_render_glyphs = False
        self._pending_position = None

        self._text_items = []
        self._old_text_items = []

        # Used by layout
        self._nlines = 0
        self._rect = Rect()

    @property
    def index(self):
        """The index in the geometry.positions buffer."""
        return self._index

    def _mark_dirty(self, *, layout=False, render_glyphs=False):
        # Trigger ._update() being called right before the next draw
        self._dirty_blocks.add(self._index)
        self._need_layout |= bool(layout) | bool(render_glyphs)
        self._need_render_glyphs |= bool(render_glyphs)

    def _update(self, text_ob):
        """Do the work to bring this block up-to-date. Be fast!"""

        # Update position?
        if self._pending_position is not None:
            positions = text_ob.geometry.positions
            positions.data[self.index] = self._pending_position
            positions.update_indices(self.index)
            self._pending_position = None

        # Reset flags
        # self._dirty_blocks.discard(self._index)  # no, Text object calls clear
        need_render_glyphs = self._need_render_glyphs
        need_layout = self._need_layout
        need_glyph_data_upload = False
        self._need_render_glyphs = False
        self._need_layout = False

        # De-allocate old item objects
        if self._old_text_items:
            need_glyph_data_upload = True
            for item in self._old_text_items:
                item.clear(text_ob)
            self._old_text_items = []

        # Quick exit
        if not (need_render_glyphs or need_layout):
            return False, False

        # Update in-use item objects
        for item in self._text_items:
            if need_render_glyphs or item.need_render_glyphs:
                item.render_glyphs(text_ob)

        # Layout
        if need_layout:
            apply_block_layout(text_ob, self)

        # Item updates, and layout, may require syncing glyph data
        for item in self._text_items:
            if item.need_sync_with_geometry:
                need_glyph_data_upload = True
                item.sync_with_geometry(text_ob, self._index)

        return need_layout, need_glyph_data_upload  # i.e. did_layout

    def _clear(self, text_ob):
        if self._old_text_items:
            for item in self._old_text_items:
                item.clear(text_ob)
            self._old_text_items = []
        for item in self._text_items:
            item.clear(text_ob)
        self._text_items = []
        self._input = None

    def set_position(self, x, y, z=0):
        self._mark_dirty()
        self._pending_position = float(x), float(y), float(z)

    def set_text(self, text: str):
        """Set the text for this TextBlock (as a string).

        This is called from ``Text.set_text()``, but can also be called
        directly. Note that in contrast to ``Text.set_text()``, setting the text
        on a TextBlock does not result in a re-layout (i.e. the bounding box is
        not updated until after the next draw).
        """

        if not isinstance(text, str):
            raise TypeError("TextBlock text should be str.")

        input = "text", text
        if input == self._input:
            return
        self._input = input
        self._mark_dirty(layout=True)

        # Split text in words
        text_parts = textmodule.tokenize_text(text)
        self._text_parts_to_items(text_parts)

    def set_markdown(self, text: str):
        """Set the markdown for this TextBlock (as a string).

        This is called from ``Text.set_markdown()``, but can also be called
        directly. Note that in contrast to ``Text.set_markdown()``, setting the
        text on a TextBlock does not result in a re-layout (i.e. the bounding
        box is not updated until after the next draw).
        """

        if not isinstance(text, str):
            raise TypeError("TextBlock markdown should be str.")

        input = "md", text
        if input == self._input:
            return
        self._input = input
        self._mark_dirty(layout=True)

        # Split text in parts using a tokenizer
        text_parts = list(textmodule.tokenize_markdown(text))

        is_newline = True
        is_bold = is_italic = 0
        max_i = len(text_parts) - 1

        for i in range(0, len(text_parts)):
            kind, text = text_parts[i]

            # Detect bullets
            if is_newline and i < max_i and text_parts[i + 1][0] == "ws":
                if text in ("*", "-"):
                    text_parts[i] = "bullet", "  •  "  # ltr and rtl compatible
                    text_parts[i + 1] = "ws", ""  # remove whitespace
                if text.startswith("#") and len(text.strip("#")) == 0:
                    header_level = len(text)
                    text_parts[i] = f"fmt:h{header_level}", text
                    text_parts[i + 1] = "ws", ""  # remove whitespace
            if kind == "nl":
                is_newline = True

            elif kind != "ws":
                is_newline = False

            # Detect bold / italics
            if kind == "stars":
                # Get what surrounding parts look like
                prev_is_wordlike = next_is_wordlike = False
                if i > 0:
                    prev_is_wordlike = text_parts[i - 1][0] != "ws"
                if i < max_i:
                    next_is_wordlike = text_parts[i + 1][0] != "ws"
                # Decide how to format
                if not prev_is_wordlike and next_is_wordlike:
                    # Might be a beginning
                    if text == "**":
                        text_parts[i] = "fmt:+b", text
                        is_bold += 1
                    elif text == "*":
                        text_parts[i] = "fmt:+i", text
                        is_italic += 1
                elif prev_is_wordlike and not next_is_wordlike:
                    # Might be an end
                    if text == "**":
                        text_parts[i] = "fmt:-b", text
                        is_bold = max(0, is_bold - 1)
                    elif text == "*":
                        text_parts[i] = "fmt:-i", text
                        is_italic = max(0, is_italic - 1)
                elif prev_is_wordlike and next_is_wordlike:
                    if text == "**":
                        if is_bold:
                            text_parts[i] = "fmt:-b", text
                            is_bold = max(0, is_bold - 1)
                        else:
                            text_parts[i] = "fmt:+b", text
                            is_bold += 1
                    elif text == "*" and is_italic:
                        if is_italic:
                            text_parts[i] = "fmt:-i", text
                            is_italic = max(0, is_italic - 1)
                        else:
                            text_parts[i] = "fmt:+i", text
                            is_italic += 1

        # Produce text items
        self._text_parts_to_items(text_parts)

    def _text_parts_to_items(self, iter_of_kind_text):
        # A TextItem represents one "word"; one thing held together during layout.
        # Multiple pieces can go into one TextItem, mostly when the word has multiple different formatting in it.
        pending_whitespace = ""
        pending_pieces = []
        new_items = []

        def flush_pieces(force=False):
            nonlocal pending_whitespace
            if pending_pieces or force:
                if self._text_items:
                    item = self._text_items.pop(0)
                else:
                    item = TextItem()
                item.set_text_pieces(tuple(pending_pieces))
                item.ws_before = pending_whitespace
                pending_pieces.clear()
                new_items.append(item)
                pending_whitespace = ""

        # In the code below, we resolve format-modifiers to an 'absolute' format.
        # These are both implementation details, but let's document them here for clarity:
        #
        # Modifiers:
        # +b: make bold
        # -b: unbold
        # +i: make italic
        # -i: make italic
        # h1 h2 h3 h4: headers
        format = {}
        bold_level = italic_level = 0

        # Process the parts to create TextItem objects
        for kind, text in iter_of_kind_text:
            if kind.startswith("fmt:"):
                modifier = kind.partition(":")[-1]
                if modifier == "+b":
                    bold_level += 1
                    format["weight"] = 300  # weight can be -250 .. 500
                elif modifier == "-b":
                    bold_level -= 1
                    if not bold_level:
                        format.pop("weight", None)
                elif modifier == "+i":
                    italic_level += 1
                    format["slant"] = 1  # slant can be -1.75 .. 2.00
                elif modifier == "-i":
                    italic_level -= 1
                    if not italic_level:
                        format.pop("slant", None)
                elif modifier.startswith("h"):
                    level = int(modifier[1:])
                    format["size"] = [1, 2.0, 1.5, 1.25][level] if level <= 3 else 1.1
            elif kind == "ws":
                flush_pieces()
                pending_whitespace += text
            elif kind == "nl":
                format.clear()
                bold_level = italic_level = 0
                flush_pieces()
                if pending_whitespace or not new_items:
                    flush_pieces(force=True)
                new_items[-1].nl_after += text
            else:
                pending_pieces.append((format.copy(), text))

        flush_pieces()
        if pending_whitespace:
            flush_pieces(force=True)

        # Store old items that need to be de-allocated
        for item in self._text_items:
            if len(item.glyph_indices):
                self._old_text_items.append(item)

        # Store new worlds
        self._text_items = new_items


class TextItem:
    """Represents one unit of text that moves as a whole to a new line when wrapped (usually a word).
    This is a low-level internal object (not public).
    """

    __slots__ = [
        "ascender",
        "atlas_indices",
        "descender",
        "direction",
        "extent",
        "formats",
        "glyph_count",
        "glyph_indices",
        "layout_offset",
        "margin_before",
        "need_render_glyphs",
        "need_sync_with_geometry",
        "nl_after",
        "offset",
        "pieces",
        "positions",
        "sizes",
        "ws_before",
    ]

    def __init__(self):
        # The text defines the arrays
        self.pieces = None

        # Whitespace attributes affect layout
        self.ws_before = ""
        self.nl_after = ""
        self.margin_before = 0  # is ws_before expressed as a float (in font units)

        # Flags to control precise updates
        self.need_render_glyphs = False
        self.need_sync_with_geometry = False

        # The text item has its own per-glyph arrays. These are copied into the geometries buffer arrays.
        self.atlas_indices = None
        self.positions = None
        self.sizes = None
        self.formats = None
        self.glyph_count = 0

        # The indices for slots in the glyph_data buffer. This value is managed by the Text object.
        self.glyph_indices = np.zeros((0,), "u4")

        # Transform info when copying into the glyph_data buffer. Set during layout.
        self.offset = (1.0, 0.0, 0.0)
        self.layout_offset = None  # used by layout as a temp var

        # Metadata
        self.extent = 0
        self.ascender = 0
        self.descender = 0
        self.direction = None

    def set_text_pieces(self, pieces):
        # The pieces arg should be [(format, text), (format, text), ...]
        if pieces != self.pieces:
            self.pieces = pieces
            self.need_render_glyphs = True
            self.margin_before = 0

    def set_offset(self, scale, dx, dy):
        offset = scale, dx, dy
        if offset != self.offset:
            self.offset = offset
            self.need_sync_with_geometry = True

    def render_glyphs(self, text_ob):
        """Update the item's arrays."""

        self.need_render_glyphs = False
        self.need_sync_with_geometry = True

        font_props = text_ob._font_props
        textengine = text_ob._text_engine

        self.margin_before = 0

        # The direction may be forced by the text object. If the text-object's direction is
        # None (default), the first piece that is word-like will determine the direction
        # for this text item.
        direction = text_ob._direction.partition("-")[0] or None

        if not self.pieces:
            self.atlas_indices = None
            self.positions = None
            self.sizes = None
            self.formats = None
            self.glyph_count = 0
            if self.ws_before:
                font = textengine.select_font(" ", font_props)[0][1]
                self.margin_before = textengine.get_ws_extent(
                    self.ws_before, font, direction
                )
            return

        # Prepare containers for array
        atlas_indices_list = []
        positions_list = []
        sizes_list = []
        formats_list = []

        # Init meta data
        extent = ascender = descender = 0
        calculate_margin_before = bool(self.ws_before)

        # Text rendering steps: font selection, shaping, glyph generation
        last_reverse_index = 0
        for format, text2 in self.pieces:
            rsize = format.get("size", 1.0)
            format_mask = encode_format(format.get("weight", 0), format.get("slant", 0))
            for text, font in textengine.select_font(text2, font_props):
                unicode_indices, positions, meta = textengine.shape_text(
                    text, font.filename, direction
                )
                atlas_indices = textengine.generate_glyph(
                    unicode_indices, font.filename
                )
                n = atlas_indices.shape[0]
                if rsize != 1.0:
                    positions *= rsize
                if extent:
                    positions[:, 0] += extent  # put pieces next to each-other
                sizes = np.full(n, rsize, "f4")
                formats = np.full(n, format_mask, "u2")
                extent = extent + meta["extent"] * rsize
                ascender = max(ascender, meta["ascender"] * rsize)
                descender = min(descender, meta["descender"] * rsize)  # is neg
                # The first piece that has actual words (not numbers or punctuation) defines direction
                if direction is None and meta["script"]:
                    direction = meta["direction"]
                # Put in list, take direction into account
                if direction in ("rtl", "btt"):
                    atlas_indices_list.insert(last_reverse_index, atlas_indices)
                    positions_list.insert(last_reverse_index, positions)
                    sizes_list.insert(last_reverse_index, sizes)
                    formats_list.insert(last_reverse_index, formats)
                else:
                    atlas_indices_list.append(atlas_indices)
                    positions_list.append(positions)
                    sizes_list.append(sizes)
                    formats_list.append(formats)
                    last_reverse_index = len(atlas_indices_list)

                # Calculate margin_before based on the font and direction of the first piece
                if calculate_margin_before:
                    calculate_margin_before = False
                    self.margin_before = textengine.get_ws_extent(
                        self.ws_before, font, direction
                    )

        # Store meta data on the item
        self.extent = extent
        self.ascender = ascender
        self.descender = descender
        self.direction = direction  # Can be None

        # Store as a single array
        if len(atlas_indices_list) == 0:
            self.atlas_indices = None
            self.positions = None
            self.sizes = None
            self.formats = None
            self.glyph_count = 0
        elif len(atlas_indices_list) == 1:
            self.atlas_indices = atlas_indices_list[0]
            self.positions = positions_list[0]
            self.sizes = sizes_list[0]
            self.formats = formats_list[0]
            self.glyph_count = len(self.positions)
        else:
            self.atlas_indices = np.concatenate(atlas_indices_list, axis=0)
            self.positions = np.concatenate(positions_list, axis=0)
            self.sizes = np.concatenate(sizes_list, axis=0)
            self.formats = np.concatenate(formats_list, axis=0)
            self.glyph_count = len(self.positions)

    def sync_with_geometry(self, text_ob, block_index):
        """Sync the item's arrays into the geometries buffers."""

        self.need_sync_with_geometry = False

        if self.glyph_count != len(self.glyph_indices):
            self._allocate_indices(text_ob)
        if self.glyph_count > 0:
            self._sync_data(text_ob, block_index)

    def _allocate_indices(self, text_ob):
        glyph_count = self.glyph_count
        glyph_indices = self.glyph_indices

        current_glyph_indices_count = len(glyph_indices)
        if glyph_count < current_glyph_indices_count:
            new_indices = glyph_indices[:glyph_count].copy()
            indices_to_free = glyph_indices[glyph_count:]
            text_ob._glyphs_deallocate(indices_to_free)
            self.glyph_indices = new_indices
        elif glyph_count > current_glyph_indices_count:
            extra_indices = text_ob._glyphs_allocate(
                glyph_count - current_glyph_indices_count
            )
            new_indices = np.empty((glyph_count,), "u4")
            new_indices[:current_glyph_indices_count] = glyph_indices
            new_indices[current_glyph_indices_count:] = extra_indices
            self.glyph_indices = new_indices

    def _sync_data(self, text_ob, block_index):
        indices = self.glyph_indices

        # TODO: I think we may optimize by combining arrays from multiple glyphs

        # Make the positioning absolute
        scale, dx, dy = self.offset
        positions = self.positions * scale + (dx, dy)
        sizes = self.sizes * scale

        # Get buffers from geometry (the getattr is a bit expensive)
        glyph_data = text_ob.geometry.glyph_data

        # Get subset and mark indices for upload
        glyph_data_array = glyph_data.data
        # glyph_data.update_indices(indices) -> doing a full upload is faster (done in _update_object)

        # Write data
        glyph_data_array["pos"][indices] = positions
        glyph_data_array["size"][indices] = sizes
        glyph_data_array["atlas_index"][indices] = self.atlas_indices
        glyph_data_array["block_index"][indices] = block_index
        glyph_data_array["format"][indices] = self.formats

    def clear(self, text_ob):
        if len(self.glyph_indices):
            text_ob._glyphs_deallocate(self.glyph_indices)
            self.glyph_indices = np.zeros((0,), "u4")


class Rect:
    __slots__ = ["bottom", "left", "right", "top"]

    def __init__(self, left=0.0, right=0.0, top=0.0, bottom=0.0):
        self.left = left
        self.right = right
        self.top = top
        self.bottom = bottom

    def __repr__(self):
        return f"<Rect({self.left:0.5g}, {self.right:0.5g}, {self.top:0.5g}, {self.bottom:0.5g})>"

    @property
    def width(self):
        return self.right - self.left

    @property
    def height(self):
        return self.top - self.bottom

    def shift(self, dx, dy):
        self.left = self.left + dx
        self.right = self.right + dx
        self.top = self.top + dy
        self.bottom = self.bottom + dy

    def get_offset_for_anchor(self, anchor, anchor_offset):
        v_anchor, _, h_anchor = anchor.partition("-")

        if h_anchor == "left":
            dx = -self.left + anchor_offset
        elif h_anchor == "right":
            dx = -self.right - anchor_offset
        else:  # center or justify
            dx = -0.5 * (self.left + self.right)

        if v_anchor == "top":
            dy = -self.top - anchor_offset
        elif v_anchor == "baseline":
            dy = -anchor_offset
        elif v_anchor == "bottom":
            dy = -self.bottom + anchor_offset
        else:  # middle
            dy = -0.5 * (self.top + self.bottom)

        return dx, dy


# ----- Layout functions


def apply_block_layout(text_ob, text_block):
    """The layout step. Updates positions and sizes to finalize the geometry."""

    items = text_block._text_items

    if not items:
        text_block._nlines = 1  # an empty line also takes the space of ine line
        text_block._rect = Rect()
        return

    # Obtain layout attributes
    font_size = text_ob._font_size
    line_height = text_ob._line_height * font_size  # like CSS
    paragraph_spacing = text_ob._paragraph_spacing * font_size
    max_width = text_ob._max_width
    anchor = text_ob._anchor
    text_align = text_ob._text_align
    text_align_last = text_ob._text_align_last
    anchor_offset = text_ob._anchor_offset

    # Get the direction to apply
    word_direction, _, line_direction = text_ob._direction.partition("-")
    line_direction = line_direction or "ttb"

    # If not overriden by the text_ob, the block direction is defined by the direction of the
    # first item that has a direction. Only items with actual words/script have a direction.
    # (Thanks to Harfbuz we can actually distinguish between script / no-script.)
    ordered_items = items
    if not word_direction:
        # Find direction
        for item in items:
            if item.direction:
                word_direction = item.direction
                break
        word_direction = word_direction or "ltr"
        # Re-order to allow subsentences in other scripts.
        # Note that in this case we only have ltr and rtl, because vertical
        # text only happens when forced by the text_ob, and then everything is the same direction.
        ordered_items = []
        i = 0
        for item in items:
            if item.direction and item.direction != word_direction:
                ordered_items.insert(i, item)
            else:
                ordered_items.append(item)
                i = len(ordered_items)

    word_direction_is_horizontal = word_direction in ("ltr", "rtl")
    line_direction_is_vertical = line_direction in ("ttb", "btt")

    # Resolve text_align_last
    if text_align_last == "auto":
        if text_align == "justify":
            text_align_last = "start"
        elif text_align == "justify_all":
            text_align_last = "justify"
        else:
            text_align_last = text_align
    if text_align == "justify_all":
        text_align = "justify"

    # Resolve text align to real directions
    if line_direction_is_vertical:
        text_align_map = {"start": "left", "end": "right"}
        if word_direction == "rtl":
            text_align_map = {"start": "right", "end": "left"}
        text_align = text_align_map.get(text_align, text_align)
        text_align_last = text_align_map.get(text_align_last, text_align_last)
    else:
        # For horizonal line direction (i.e. vertical word-direction)
        # we don't support text align.
        text_align = text_align_last = "left" if line_direction == "ltr" else "right"

    # If we wrote the above correctly, we have a clear subset of alignment vars
    assert text_align in ("left", "right", "center", "justify")
    assert text_align_last in ("left", "right", "center", "justify")

    # Prepare

    # The offset is used to track the position as we wrap the text
    offset = [0, 0]

    # The current rect represents the bounding box of each line, relative to the line.
    # Vertically, point zero is at the baseline, and the rect spans from ascender (top, positive) to descender (bottom, negative).
    current_rect = Rect()  # left, right, top, bottom

    # The current line holds the text items for the line being processed.
    current_line = []

    lines_rects = []  # List[ Tuple[ List[TextItem], Rect] ]

    def make_new_line(n_new_lines=1, n_new_paragraphs=0):
        nonlocal current_line, current_rect
        # Update position
        skip = n_new_lines * line_height + n_new_paragraphs * paragraph_spacing
        if line_direction == "ttb":
            offset[1] -= skip
            offset[0] = 0
        elif line_direction == "btt":
            offset[1] += skip
            offset[0] = 0
        elif line_direction == "rtl":
            offset[1] = 0
            offset[0] -= skip
        elif line_direction == "ltr":
            offset[1] = 0
            offset[0] += skip
        # Add the line
        if current_line:
            lines_rects.append((current_line, current_rect))
            current_line = []
            current_rect = Rect()

    # Resolve position and sizes
    for item in ordered_items:
        # Get item width and determine if we need a new line
        apply_whitespace_margin = True
        if max_width > 0 and current_line:
            if word_direction_is_horizontal:
                current_width = current_rect.width
            else:
                current_width = current_rect.height

            item_width = (item.margin_before + item.extent) * font_size
            if current_width + item_width > max_width:
                make_new_line()
                apply_whitespace_margin = False

        if word_direction == "ltr":
            # Update offset
            if apply_whitespace_margin:
                offset[0] += item.margin_before * font_size
            item.layout_offset = tuple(offset)
            offset[0] += item.extent * font_size
            # Update rect
            current_rect.left = 0
            current_rect.right = offset[0]
            current_rect.top = max(current_rect.top, item.ascender * font_size)
            current_rect.bottom = min(current_rect.bottom, item.descender * font_size)
        elif word_direction == "rtl":
            # Update offset
            if apply_whitespace_margin:
                offset[0] -= item.margin_before * font_size
            offset[0] -= item.extent * font_size
            item.layout_offset = tuple(offset)
            # Update rect
            current_rect.left = offset[0]
            current_rect.right = 0
            current_rect.top = max(current_rect.top, item.ascender * font_size)
            current_rect.bottom = min(current_rect.bottom, item.descender * font_size)
        elif word_direction == "ttb":
            # Update offset
            if apply_whitespace_margin:
                offset[1] -= item.margin_before * font_size
            item.layout_offset = tuple(offset)
            offset[1] -= item.extent * font_size
            # Update rect
            current_rect.top = 0
            current_rect.bottom = offset[1]
            current_rect.left = item.descender * font_size
            current_rect.right = item.ascender * font_size
        elif word_direction == "btt":
            # Update offset
            if apply_whitespace_margin:
                offset[1] += item.margin_before * font_size
            offset[1] += item.extent * font_size
            item.layout_offset = tuple(offset)
            # Update rect
            current_rect.top = offset[1]
            current_rect.bottom = 0
            current_rect.left = item.descender * font_size
            current_rect.right = item.ascender * font_size

        current_line.append(item)

        # The item can have newlines too. Does not happen when using text_ob.set_text(),
        # but can happen when using TextBlock.set_text().
        if item.nl_after:
            make_new_line(len(item.nl_after), 1)

    # Properly end the loop
    if current_line:
        make_new_line()

    # Calculate block rect. The top is positive, the bottom is negative (descender).
    block_rect = Rect()
    for line, rect in lines_rects:
        # For rtl, align each line so left is at the origin
        if word_direction == "rtl":
            shift = -rect.left
            rect.shift(shift, 0)
            for item in line:
                layout_offset = item.layout_offset
                item.layout_offset = layout_offset[0] + shift, layout_offset[1]
        # Aggregate
        if line_direction_is_vertical:
            block_rect.left = min(block_rect.left, rect.left)
            block_rect.right = max(block_rect.right, rect.right)
        else:
            block_rect.top = max(block_rect.top, rect.top)
            block_rect.bottom = min(block_rect.bottom, rect.bottom)

    first_rect, last_rect = lines_rects[0][1], lines_rects[-1][1]
    if line_direction == "ttb":
        block_rect.top = first_rect.top
        block_rect.bottom = offset[1] + line_height + last_rect.bottom
    elif line_direction == "btt":
        block_rect.top = offset[1] - line_height + last_rect.top
        block_rect.bottom = first_rect.bottom
    elif line_direction == "rtl":
        block_rect.left = offset[0] + line_height + last_rect.left
        block_rect.right = first_rect.right
    elif line_direction == "ltr":
        block_rect.left = first_rect.left
        block_rect.right = offset[0] - line_height + last_rect.right

    if text_align == "justify" or text_align_last == "justify":
        if max_width > 0:
            block_rect.right = max_width
        else:
            # Within a block, support justify without max-width
            max_width = block_rect.right

    # Determine horizontal anchor
    if text_ob.is_multi:
        # Full layout done here, including anchoring.
        anchor_offset_x, anchor_offset_y = block_rect.get_offset_for_anchor(
            anchor, anchor_offset
        )
    else:
        # If the text_ob does its layout, it's far easier to *not* do the anchoring here,
        # except to anchor according to text alignment.
        anchor_offset_x, anchor_offset_y = block_rect.get_offset_for_anchor(
            f"baseline-{text_align}", 0
        )

    # Shift block rect for anchoring
    block_rect.shift(anchor_offset_x, anchor_offset_y)

    # Align the text, i.e. shift individual lines so they fit inside the block rect according to the current alignment

    num_lines = len(lines_rects)
    align = text_align
    for i, (line, rect) in enumerate(lines_rects):
        if i == num_lines - 1:
            align = text_align_last

        line_length = rect.right - rect.left
        block_length = block_rect.right - block_rect.left

        extra_space_per_word = 0
        if not word_direction_is_horizontal:
            line_offset_x = 0
        elif align == "justify":
            line_offset_x = 0
            length_to_add = max_width - line_length
            nwords = len(line)
            if nwords > 1:
                extra_space_per_word = length_to_add / (nwords - 1)
        elif align == "center":
            line_offset_x = 0.5 * block_length - 0.5 * line_length
        elif align == "right":
            line_offset_x = block_length - line_length
        else:  # elif align == "left":
            line_offset_x = 0

        for j, item in enumerate(line):
            if word_direction == "rtl":
                j = len(line) - j - 1
            dx = (
                item.layout_offset[0]
                + anchor_offset_x
                + line_offset_x
                + j * extra_space_per_word
            )
            dy = item.layout_offset[1] + anchor_offset_y
            item.set_offset(font_size, dx, dy)

    # Update block's rect. Used by the final layout and to calculate bounding boxes.
    text_block._rect = block_rect
    text_block._nlines = num_lines


def apply_high_level_layout(text_ob):
    text_blocks = text_ob._text_blocks

    if not text_blocks:
        text_ob._aabb = np.zeros((2, 3), "f4")
        return

    font_size = text_ob._font_size
    anchor = text_ob._anchor
    anchor_offset = text_ob._anchor_offset
    line_height = text_ob._line_height * font_size  # like CSS
    par_spacing = text_ob._paragraph_spacing * font_size

    # Get line direction.
    line_direction = text_ob._direction.partition("-")[2] or "ttb"
    line_direction_is_vertical = line_direction in ("ttb", "btt")

    # Calculate offsets to put the blocks beneath each-other, as well as the full rect.
    # Note that the distance between anchor points is independent on the anchor-mode.
    offsets = np.zeros((len(text_blocks),), "f4")
    offset = 0
    total_rect = Rect()
    if line_direction == "ttb":
        for i, block in enumerate(text_blocks):
            rect = block._rect
            offsets[i] = offset
            offset -= max(block._nlines * line_height, 0.5 * rect.height) + par_spacing
            total_rect.left = min(total_rect.left, rect.left)
            total_rect.right = max(total_rect.right, rect.right)
        total_rect.top = text_blocks[0]._rect.top
        total_rect.bottom = offsets[-1] + text_blocks[-1]._rect.bottom
    elif line_direction == "btt":
        for i, block in enumerate(text_blocks):
            rect = block._rect
            offsets[i] = offset
            offset += max(block._nlines * line_height, 0.5 * rect.height) + par_spacing
            total_rect.left = min(total_rect.left, rect.left)
            total_rect.right = max(total_rect.right, rect.right)
        total_rect.top = offsets[-1] + text_blocks[-1]._rect.top
        total_rect.bottom = text_blocks[0]._rect.bottom
    elif line_direction == "rtl":
        for i, block in enumerate(text_blocks):
            rect = block._rect
            offsets[i] = offset
            offset -= max(block._nlines * line_height, 0.5 * rect.width) + par_spacing
            total_rect.bottom = min(total_rect.bottom, rect.bottom)
            total_rect.top = max(total_rect.top, rect.top)
        total_rect.left = offsets[-1] + text_blocks[-1]._rect.left
        total_rect.right = text_blocks[0]._rect.right
    elif line_direction == "ltr":
        for i, block in enumerate(text_blocks):
            rect = block._rect
            offsets[i] = offset
            offset += max(block._nlines * line_height, 0.5 * rect.width) + par_spacing
            total_rect.bottom = min(total_rect.bottom, rect.bottom)
            total_rect.top = max(total_rect.top, rect.top)
        total_rect.left = text_blocks[0]._rect.left
        total_rect.right = offsets[-1] + text_blocks[-1]._rect.right

    # Get anchor offset
    # Note that the anchoring is dead-simple because the blocks are anchoring based on text_align.
    anchor_offset_x, anchor_offset_y = total_rect.get_offset_for_anchor(
        anchor, anchor_offset
    )

    # Shift bounding box rect, and store for geomerty bounding box.
    # Note how we swap top and bottom here, because bottom has smaller values than top.
    total_rect.shift(anchor_offset_x, anchor_offset_y)

    # Update positions
    positions = text_ob.geometry.positions
    if line_direction_is_vertical:
        positions.data[: len(offsets), 0] = anchor_offset_x
        positions.data[: len(offsets), 1] = offsets + anchor_offset_y
    else:
        positions.data[: len(offsets), 0] = offsets + anchor_offset_x
        positions.data[: len(offsets), 1] = anchor_offset_y

    # Update full positions buffer to avoid overhead of chunking logic. Measurably faster.
    positions.update_full()
    # positions.update_range(0, len(offsets))

    return total_rect



"""World Objects and Events.

.. currentmodule:: pygfx.objects

.. rubric:: Objects
.. autosummary::
    :toctree: objects/
    :template: ../_templates/custom_layout.rst

    WorldObject
    Group
    Scene
    Background
    Grid
    Points
    Line
    Mesh
    Image
    Volume
    Text
    MultiText
    TextBlock
    Ruler
    InstancedMesh
    InstancedLine
    Light
    PointLight
    DirectionalLight
    AmbientLight
    SpotLight
    LightShadow
    DirectionalLightShadow
    SpotLightShadow
    PointLightShadow
    Bone
    Skeleton
    SkinnedMesh

.. rubric:: Events
.. autosummary::
    :toctree: objects/
    :template: ../_templates/custom_layout.rst

    Event
    EventTarget
    EventType
    PointerEvent
    KeyboardEvent
    RootEventHandler
    WheelEvent
    WindowEvent

"""

# ruff: noqa: F401

from ._base import WorldObject
from ._events import (
    Event,
    EventTarget,
    EventType,
    PointerEvent,
    KeyboardEvent,
    RootEventHandler,
    WheelEvent,
    WindowEvent,
)
from ._more import (
    Group,
    Scene,
    Background,
    Grid,
    Points,
    Line,
    Mesh,
    Image,
    Volume,
)
from ._ruler import Ruler
from ._text import Text, MultiText, TextBlock
from ._instanced import InstancedMesh, InstancedLine
from ._lights import Light, PointLight, DirectionalLight, AmbientLight, SpotLight
from ._lights import (
    LightShadow,
    DirectionalLightShadow,
    SpotLightShadow,
    PointLightShadow,
)
from ._skins import Bone, Skeleton, SkinnedMesh



import numpy as np

from ._base import Geometry


# Note that we keep this function separate, because its used by other geometry-generating functions.


def generate_plane(width, height, width_segments, height_segments):
    nx, ny = width_segments + 1, height_segments + 1

    x = np.linspace(-width / 2, width / 2, nx, dtype=np.float32)
    y = np.linspace(-height / 2, height / 2, ny, dtype=np.float32)
    xx, yy = np.meshgrid(x, y)
    xx, yy = xx.flatten(), yy.flatten()
    positions = np.column_stack([xx, yy, np.zeros_like(xx)])

    dim = np.array([width, height], dtype=np.float32)
    texcoords = (positions[..., :2] + dim / 2) / dim
    texcoords[..., 1] = 1 - texcoords[..., 1]

    # the amount of vertices
    indices = np.arange(ny * nx, dtype=np.uint32).reshape((ny, nx))
    # for every panel (height_segments, width_segments) there is a quad (2, 3)
    index = np.empty((height_segments, width_segments, 2, 3), dtype=np.uint32)
    # create a grid of initial indices for the panels
    index[:, :, 0, 0] = indices[
        np.arange(height_segments)[:, None], np.arange(width_segments)[None, :]
    ]
    # the remainder of the indices for every panel are relative
    index[:, :, 0, 1] = index[:, :, 0, 0] + 1
    index[:, :, 0, 2] = index[:, :, 0, 0] + nx
    index[:, :, 1, 0] = index[:, :, 0, 0] + nx + 1
    index[:, :, 1, 1] = index[:, :, 1, 0] - 1
    index[:, :, 1, 2] = index[:, :, 1, 0] - nx

    normals = np.tile(np.array([0, 0, 1], dtype=np.float32), (ny * nx, 1))

    return positions, normals, texcoords, index.reshape((-1, 3))


def plane_geometry(width=1, height=1, width_segments=1, height_segments=1):
    """Generate a plane.

    Creates a flat (2D) rectangle in the local xy-plane that has its center at
    local origin. The plane may be subdivided into segments along the x- or
    y-axis respectively.

    Parameters
    ----------
    width : float
        The plane's width measured along the x-axis.
    height : float
        The plane's width measured along the y-axis.
    width_segments : int
        The number of evenly spaced segments along the x-axis into which the
        plane should be divided.
    height_segments : int
        The number of evenly spaced segments along the y-axis into which the
        plane should be divided.

    Returns
    -------
    plane : Geometry
        A geometry object representing the requested plane.
        Mathematically, it is an open orientable manifold.

    """

    positions, normals, texcoords, indices = generate_plane(
        width, height, width_segments, height_segments
    )

    return Geometry(
        indices=indices.reshape((-1, 3)),
        positions=positions,
        normals=normals,
        texcoords=texcoords,
    )


def mobius_strip_geometry(radius=1, width=0.5, strip_segments=64, stitch=True):
    """Generate a Möbius strip.

    The Möbios strip is a surface that can be formed by attaching the
    ends of a strip of paper together with a half-twist.

    Parameters
    ----------
    radius : float
        The radius of the circe along which the strip is positioned.
    width : float
        The width of the strip.
    strip_segments : int
        The number of evenly spaced segments along the strip.
    stitch : bool
        Whether to stitch the ends together to form a closed loop. If False (default)
        the strip's ends meet, but are not actually attached to one another.

    Returns
    -------
    strip : Geometry
        A geometry object representing the Möbius strip.
        Mathematically, it is an open manifold, that is only orientable
        if ``stitch`` is False.

    """

    # Check/convert inputs
    n_strip = int(strip_segments)
    radius = float(radius)
    width = float(width)
    stitch = bool(stitch)

    n_verts = n_strip + (0 if stitch else 1)

    # Create base vectors
    t = np.linspace(0, np.pi * 2, n_verts, endpoint=not stitch, dtype=np.float32)
    u = np.linspace(-width / 2, width / 2, 2, endpoint=True, dtype=np.float32)
    u, t = np.meshgrid(u, t)

    # The math
    x = radius * np.cos(t) + u * np.sin(0.5 * t)
    z = radius * np.sin(t)
    y = u * np.cos(0.5 * t)

    # Define connectivity
    indices = np.array([0, 1, 2, 2, 1, 3], np.uint32)
    indices = np.tile(indices, (n_strip, 1, 1))
    indices += np.arange(0, n_strip * 2, 2, dtype=np.uint32).reshape(n_strip, 1, 1)

    # Stitch the ends
    if stitch:
        indices[-1, :, (2, 3, 5)] = np.array((1, 1, 0)).reshape(3, 1)

    return Geometry(
        indices=indices.reshape((-1, 3)),
        positions=np.column_stack([x.flat, y.flat, z.flat]),
    )



from ._base import Material
from ..resources import Texture, TextureMap
from ..utils import unpack_bitfield, Color, assert_type
from ..utils.enums import ColorMode, CoordSpace


class LineMaterial(Material):
    """Basic line material.

    Parameters
    ----------
    thickness : float
        The line thickness expressed in logical pixels. Default 2.0.
    thickness_space : str | CoordSpace
        The coordinate space in which the thickness is expressed ('screen', 'world', 'model'). Default 'screen'.
    color : Color
        The uniform color of the line (used depending on the ``color_mode``).
    color_mode : str | ColorMode
        The mode by which the line is coloured. Default 'auto'.
    map : TextureMap | Texture
        The texture map specifying the color for each texture coordinate. Optional.
    dash_pattern : tuple
        The pattern of the dash, e.g. `[2, 3]`. See `dash_pattern` docs for details. Defaults to an empty tuple, i.e. no dashing.
    dash_offset : float
        The offset into the dash phase. Default 0.0.
    loop : bool
        Whether the line's end should be connected. Default False.
    aa : bool
        Whether the line is anti-aliased in the shader. Default False.
    kwargs : Any
        Additional kwargs will be passed to the :class:`material base class <pygfx.Material>`.
    """

    uniform_type = dict(
        Material.uniform_type,
        color="4xf4",
        thickness="f4",
        dash_offset="f4",
    )

    def __init__(
        self,
        thickness=2.0,
        thickness_space="screen",
        *,
        color=(1, 1, 1, 1),
        color_mode="auto",
        map=None,
        dash_pattern=(),
        dash_offset=0,
        loop=False,
        aa=False,
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.thickness = thickness
        self.thickness_space = thickness_space
        self.color = color
        self.color_mode = color_mode
        self.map = map
        self.dash_pattern = dash_pattern
        self.dash_offset = dash_offset
        self.loop = loop
        self.aa = aa

    def _wgpu_get_pick_info(self, pick_value):
        # This should match with the shader
        values = unpack_bitfield(pick_value, wobject_id=20, index=26, coord=18)
        return {
            "vertex_index": values["index"],
            "segment_coord": (values["coord"] - 100000) / 100000.0,
        }

    @property
    def color(self):
        """The uniform color of the line."""
        return Color(self.uniform_buffer.data["color"])

    @color.setter
    def color(self, color):
        color = Color(color)
        self.uniform_buffer.data["color"] = color
        self.uniform_buffer.update_full()

    @property
    def aa(self):
        """Whether the line's edges are anti-aliased.

        Aliasing gives prettier results by producing semi-transparent fragments
        at the edges. Lines thinner than one physical pixel are also diminished
        by making them more transparent.

        However, because semi-transparent fragments are introduced, artifacts
        may occur if certain cases. For the same reason, aa only works for the
        "blended" and "weighted" alpha methods.

        Note that by default, pygfx already uses SSAA and/or PPAA to anti-alias
        the total renderered result. Line-based aa is an *additional* visual
        improvement.
        """
        return self._store.aa

    @aa.setter
    def aa(self, aa):
        self._store.aa = bool(aa)

    @property
    def _gfx_effective_aa(self):
        aa_able_methods = ("blended", "weighted")
        return self._store.aa and self.alpha_method in aa_able_methods

    @property
    def color_mode(self):
        """The way that color is applied to the line.

        See :obj:`pygfx.utils.enums.ColorMode`:
        """
        return self._store.color_mode

    @color_mode.setter
    def color_mode(self, value):
        value = value or "auto"
        if value not in ColorMode:
            raise ValueError(
                f"LineMaterial.color_mode must be a string in {ColorMode}, not {value!r}"
            )
        self._store.color_mode = value

    @property
    def vertex_colors(self):
        return self.color_mode == ColorMode.vertex

    @vertex_colors.setter
    def vertex_colors(self, value):
        raise DeprecationWarning(
            "vertex_colors is deprecated, use ``color_mode='vertex'``"
        )

    @property
    def thickness(self):
        """The line thickness.

        The interpretation depends on `thickness_space`. By default it is in logical
        pixels, but it can also be in world or model coordinates.
        """
        return float(self.uniform_buffer.data["thickness"])

    @thickness.setter
    def thickness(self, thickness):
        self.uniform_buffer.data["thickness"] = max(0.0, float(thickness))
        self.uniform_buffer.update_full()

    @property
    def thickness_space(self):
        """The coordinate space in which the thickness (and dash_pattern) are expressed.

        See :obj:`pygfx.utils.enums.CoordSpace`:
        """
        return self._store.thickness_space

    @thickness_space.setter
    def thickness_space(self, value):
        value = value or "screen"
        if value not in CoordSpace:
            raise ValueError(
                f"LineMaterial.thickness_space must be a string in {CoordSpace}, not {value!r}"
            )
        self._store.thickness_space = value

    @property
    def map(self):
        """The texture map specifying the color for each texture coordinate.

        Can be None. The dimensionality of the map can be 1D, 2D or 3D, but
        should match the number of columns in the geometry's texcoords.
        """
        return self._store.map

    @map.setter
    def map(self, map):
        assert_type("map", map, None, Texture, TextureMap)
        if isinstance(map, Texture):
            map = TextureMap(map)
        self._store.map = map

    @property
    def dash_pattern(self):
        """The dash pattern.

        A sequence of floats describing the length of strokes and gaps. The
        length of the sequence must be an even number. Setting to None or the
        empty tuple means no dashing.

        For example, (5, 2, 1, 2) describes a a stroke of 5 units, a gap of 2,
        then a short stroke of 1, and another gap of 2. Units are relative to
        the line thickness (and therefore `thickness_space` also applies to  the
        `dash_pattern`).
        """
        return self._store.dash_pattern

    @dash_pattern.setter
    def dash_pattern(self, value):
        if value is None:
            value = ()
        if not isinstance(value, (tuple, list)):
            raise TypeError(
                "Line dash_pattern must be a sequence of floats, not '{value}'"
            )
        if len(value) % 2:
            raise ValueError("Line dash_pattern must have an even number of elements.")
        self._store.dash_pattern = tuple(max(0.0, float(v)) for v in value)

    @property
    def dash_offset(self):
        """The offset into the dash cycle to start drawing at, i.e. the phase."""
        return float(self.uniform_buffer.data["dash_offset"])

    @dash_offset.setter
    def dash_offset(self, value):
        self.uniform_buffer.data["dash_offset"] = float(value)
        self.uniform_buffer.update_full()

    @property
    def loop(self) -> bool:
        """Whether the line's ends should be connected.

        If set to True, the end of the line is connected to its beginning, in
        such a way there is no overlap (which would otherwise be visible for
        semi-transparent lines). When the line consists of multiple pieces
        separated by nan-positions, each line-piece is considered a loop.
        """
        return self._store.loop

    @loop.setter
    def loop(self, loop: bool):
        self._store.loop = bool(loop)


class LineDebugMaterial(LineMaterial):
    """Line debug material.

    A material that renders the triangles that the line is made up off.
    """

    pass


class LineSegmentMaterial(LineMaterial):
    """Line segment material.

    A material that renders line segments between each two subsequent points.
    """


class LineInfiniteSegmentMaterial(LineSegmentMaterial):
    """Infinite line segment material.

    A material that renders infenitely long line segments between each two
    subsequent points. The end-points of each segment are displaced (along the
    vector defined by the two points) such that the points are at the edge of
    the viewport. Other than that, dashing, vertex colors, etc. should work as
    expected (interpolating between the points that are now on the viewport edge).

    Parameters
    ----------
    start_is_infinite : bool
        Whether start of each segment is made infinitely long. Default True.
    end_is_infinite : bool
        Whether end of each segment is made infinitely long. Default True.
    """

    def __init__(self, start_is_infinite=True, end_is_infinite=True, **kwargs):
        super().__init__(**kwargs)
        self.start_is_infinite = start_is_infinite
        self.end_is_infinite = end_is_infinite

    @property
    def start_is_infinite(self):
        """Whether start of each segment is made infinitely long."""
        return self._store.start_is_infinite

    @start_is_infinite.setter
    def start_is_infinite(self, value):
        self._store.start_is_infinite = bool(value)

    @property
    def end_is_infinite(self):
        """Whether end of each segment is made infinitely long."""
        return self._store.end_is_infinite

    @end_is_infinite.setter
    def end_is_infinite(self, value):
        self._store.end_is_infinite = bool(value)


class LineArrowMaterial(LineSegmentMaterial):
    """Arrow (vector) line material.

    A material that renders line segments that look like little arrows.
    """


class LineThinMaterial(LineMaterial):
    """Thin line material.

    A simple line, drawn with line_strip primitives that has a thickness
    of one physical pixel. Thickness, dashing, and aa are ignored.

    While you typically don't want to use this in your application (its
    width is inconsistent and looks *very* thin on HiDPI monitors), it can be
    useful for debugging as it is more performant than other line materials.

    """


class LineThinSegmentMaterial(LineMaterial):
    """Thin line segment material.

    Simple line segments, drawn with line primitives that has a thickness
    of one physical pixel. Thickness, dashing, and aa are ignored.

    While you typically don't want to use this in your application (its
    width is inconsistent and looks *very* thin on HiDPI monitors), it can be
    useful for debugging as it is more performant than other line materials.

    """
